from fastapi_users.exceptions import UserAlreadyExists
from cognee.modules.users.exceptions import TenantNotFoundError
from cognee.infrastructure.databases.relational import get_relational_engine
from cognee.modules.users.get_user_manager import get_user_manager_context
from cognee.modules.users.get_user_db import get_user_db_context
from cognee.modules.users.models.User import UserCreate
from cognee.modules.users.models.Tenant import Tenant

from sqlalchemy import select
from typing import Optional


async def create_user(
    email: str,
    password: str,
    tenant_id: Optional[str] = None,
    is_superuser: bool = False,
    is_active: bool = True,
    is_verified: bool = False,
    auto_login: bool = False,
):
    try:
        relational_engine = get_relational_engine()

        async with relational_engine.get_async_session() as session:
            async with get_user_db_context(session) as user_db:
                async with get_user_manager_context(user_db) as user_manager:
                    if tenant_id:
                        # Check if the tenant already exists
                        result = await session.execute(select(Tenant).where(Tenant.id == tenant_id))
                        tenant = result.scalars().first()
                        if not tenant:
                            raise TenantNotFoundError

                        user = await user_manager.create(
                            UserCreate(
                                email=email,
                                password=password,
                                tenant_id=tenant.id,
                                is_superuser=is_superuser,
                                is_active=is_active,
                                is_verified=is_verified,
                            )
                        )
                    else:
                        user = await user_manager.create(
                            UserCreate(
                                email=email,
                                password=password,
                                is_superuser=is_superuser,
                                is_active=is_active,
                                is_verified=is_verified,
                            )
                        )

                    if auto_login:
                        await session.refresh(user)

                    return user
    except UserAlreadyExists as error:
        print(f"User {email} already exists")
        raise error



from enum import Enum
from typing import Type

from cognee.eval_framework.benchmark_adapters.hotpot_qa_adapter import HotpotQAAdapter
from cognee.eval_framework.benchmark_adapters.musique_adapter import MusiqueQAAdapter
from cognee.eval_framework.benchmark_adapters.dummy_adapter import DummyAdapter
from cognee.eval_framework.benchmark_adapters.twowikimultihop_adapter import TwoWikiMultihopAdapter


class BenchmarkAdapter(Enum):
    DUMMY = ("Dummy", DummyAdapter)
    HOTPOTQA = ("HotPotQA", HotpotQAAdapter)
    MUSIQUE = ("Musique", MusiqueQAAdapter)
    TWOWIKIMULTIHOP = ("TwoWikiMultiHop", TwoWikiMultihopAdapter)

    def __new__(cls, adapter_name: str, adapter_class: Type):
        obj = object.__new__(cls)
        obj._value_ = adapter_name
        obj.adapter_class = adapter_class
        return obj

    def __str__(self):
        return self.value



import pytest
from cognee.eval_framework.corpus_builder.corpus_builder_executor import CorpusBuilderExecutor
from cognee.infrastructure.databases.graph import get_graph_engine
from unittest.mock import AsyncMock, patch

benchmark_options = ["HotPotQA", "Dummy", "TwoWikiMultiHop"]


@pytest.mark.parametrize("benchmark", benchmark_options)
def test_corpus_builder_load_corpus(benchmark):
    limit = 2
    corpus_builder = CorpusBuilderExecutor(benchmark, "Default")
    raw_corpus, questions = corpus_builder.load_corpus(limit=limit)
    assert len(raw_corpus) > 0, f"Corpus builder loads empty corpus for {benchmark}"
    assert len(questions) <= 2, (
        f"Corpus builder loads {len(questions)} for {benchmark} when limit is {limit}"
    )


@pytest.mark.asyncio
@pytest.mark.parametrize("benchmark", benchmark_options)
@patch.object(CorpusBuilderExecutor, "run_cognee", new_callable=AsyncMock)
async def test_corpus_builder_build_corpus(mock_run_cognee, benchmark):
    limit = 2
    corpus_builder = CorpusBuilderExecutor(benchmark, "Default")
    questions = await corpus_builder.build_corpus(limit=limit)
    assert len(questions) <= 2, (
        f"Corpus builder loads {len(questions)} for {benchmark} when limit is {limit}"
    )



"""LLM Interface"""

from typing import Type, Protocol
from abc import abstractmethod
from pydantic import BaseModel
from cognee.infrastructure.llm.prompts import read_query_prompt


class LLMInterface(Protocol):
    """LLM Interface"""

    @abstractmethod
    async def acreate_structured_output(
        self, text_input: str, system_prompt: str, response_model: Type[BaseModel]
    ) -> BaseModel:
        """To get structured output, import/call this function"""
        raise NotImplementedError

    def show_prompt(self, text_input: str, system_prompt: str) -> str:
        """Format and display the prompt for a user query."""
        if not text_input:
            text_input = "No user input provided."
        if not system_prompt:
            raise ValueError("No system prompt path provided.")
        system_prompt = read_query_prompt(system_prompt)

        formatted_prompt = f"""System Prompt:\n{system_prompt}\n\nUser Input:\n{text_input}\n"""

        return formatted_prompt



from uuid import UUID
from sqlalchemy import select
from cognee.infrastructure.databases.relational import get_relational_engine
from cognee.modules.data.models import Dataset, DatasetData, Data
from ...models import ACL, Permission


async def get_document_ids_for_user(user_id: UUID, datasets: list[str] = None) -> list[str]:
    db_engine = get_relational_engine()

    async with db_engine.get_async_session() as session:
        async with session.begin():
            document_ids = (
                await session.scalars(
                    select(Data.id)
                    .join(ACL.data)
                    .join(ACL.permission)
                    .where(
                        ACL.principal_id == user_id,
                        Permission.name == "read",
                    )
                )
            ).all()

            if datasets:
                documents_ids_in_dataset = set()
                # If datasets are specified filter out documents that aren't part of the specified datasets
                for dataset in datasets:
                    # Find dataset id for dataset element
                    dataset_id = (
                        await session.scalars(
                            select(Dataset.id).where(
                                Dataset.name == dataset,
                                Dataset.owner_id == user_id,
                            )
                        )
                    ).one_or_none()

                    # Check which documents are connected to this dataset
                    for document_id in document_ids:
                        data_id = (
                            await session.scalars(
                                select(DatasetData.data_id).where(
                                    DatasetData.dataset_id == dataset_id,
                                    DatasetData.data_id == document_id,
                                )
                            )
                        ).one_or_none()

                        # If document is related to dataset added it to return value
                        if data_id:
                            documents_ids_in_dataset.add(document_id)
                return list(documents_ids_in_dataset)
            return document_ids



import os
import pathlib
import cognee
from cognee.modules.search.operations import get_history
from cognee.modules.users.methods import get_default_user
from cognee.shared.logging_utils import get_logger
from cognee.modules.search.types import SearchType
from cognee.shared.utils import render_graph
from cognee.low_level import DataPoint

logger = get_logger()


async def main():
    data_directory_path = str(
        pathlib.Path(
            os.path.join(pathlib.Path(__file__).parent, ".data_storage/test_custom_model")
        ).resolve()
    )
    cognee.config.data_root_directory(data_directory_path)
    cognee_directory_path = str(
        pathlib.Path(
            os.path.join(pathlib.Path(__file__).parent, ".cognee_system/test_custom_model")
        ).resolve()
    )
    cognee.config.system_root_directory(cognee_directory_path)

    await cognee.prune.prune_data()
    await cognee.prune.prune_system(metadata=True)

    # Define a custom graph model for programming languages.
    class FieldType(DataPoint):
        name: str = "Field"
        metadata: dict = {"index_fields": ["name"]}

    class Field(DataPoint):
        name: str
        is_type: FieldType
        metadata: dict = {"index_fields": ["name"]}

    class ProgrammingLanguageType(DataPoint):
        name: str = "Programming Language"
        metadata: dict = {"index_fields": ["name"]}

    class ProgrammingLanguage(DataPoint):
        name: str
        used_in: list[Field] = []
        is_type: ProgrammingLanguageType
        metadata: dict = {"index_fields": ["name"]}

    text = (
        "Python is an interpreted, high-level, general-purpose programming language. It was created by Guido van Rossum and first released in 1991. "
        + "Python is widely used in data analysis, web development, and machine learning."
    )

    await cognee.add(text)

    await cognee.cognify(graph_model=ProgrammingLanguage)

    url = await render_graph()
    print(f"Graphistry URL: {url}")

    graph_file_path = str(
        pathlib.Path(
            os.path.join(
                pathlib.Path(__file__).parent,
                ".artifacts/test_custom_model/graph_visualization.html",
            )
        ).resolve()
    )
    await cognee.visualize_graph(graph_file_path)

    # Completion query that uses graph data to form context.
    completion = await cognee.search(SearchType.GRAPH_COMPLETION, "What is python?")
    assert len(completion) != 0, "Graph completion search didn't return any result."
    print("Graph completion result is:")
    print(completion)

    # Completion query that uses document chunks to form context.
    completion = await cognee.search(SearchType.RAG_COMPLETION, "What is Python?")
    assert len(completion) != 0, "Completion search didn't return any result."
    print("Completion result is:")
    print(completion)

    # Query all summaries related to query.
    summaries = await cognee.search(SearchType.SUMMARIES, "Python")
    assert len(summaries) != 0, "Summaries search didn't return any results."
    print("Summary results are:")
    for summary in summaries:
        print(summary)

    chunks = await cognee.search(SearchType.CHUNKS, query_text="Python")
    assert len(chunks) != 0, "Chunks search didn't return any results."
    print("Chunk results are:")
    for chunk in chunks:
        print(chunk)

    user = await get_default_user()
    history = await get_history(user.id)

    assert len(history) == 8, "Search history is not correct."


if __name__ == "__main__":
    import asyncio

    asyncio.run(main(), debug=True)



import pytest
from cognee.tasks.entity_completion.entity_extractors.regex_entity_extractor import (
    RegexEntityExtractor,
)


@pytest.fixture
def regex_extractor():
    """Create a RegexEntityExtractor instance for testing."""
    return RegexEntityExtractor()


@pytest.mark.asyncio
async def test_extract_emails(regex_extractor):
    """Test extraction of email addresses."""
    text = "Contact us at support@example.com or sales@company.co.uk for more information."
    entities = await regex_extractor.extract_entities(text)

    # Filter only EMAIL entities
    email_entities = [e for e in entities if e.is_a.name == "EMAIL"]

    assert len(email_entities) == 2
    assert "support@example.com" in [e.name for e in email_entities]
    assert "sales@company.co.uk" in [e.name for e in email_entities]


@pytest.mark.asyncio
async def test_extract_phone_numbers(regex_extractor):
    """Test extraction of phone numbers."""
    text = "Call us at +1-555-123-4567 or 020 7946 0958 for support."
    entities = await regex_extractor.extract_entities(text)

    # Filter only PHONE entities
    phone_entities = [e for e in entities if e.is_a.name == "PHONE"]

    assert len(phone_entities) == 2
    assert "+1-555-123-4567" in [e.name for e in phone_entities]
    assert "020 7946 0958" in [e.name for e in phone_entities]


@pytest.mark.asyncio
async def test_extract_urls(regex_extractor):
    """Test extraction of URLs."""
    text = "Visit our website at https://www.example.com or http://docs.example.org/help for more information."
    entities = await regex_extractor.extract_entities(text)

    # Filter only URL entities
    url_entities = [e for e in entities if e.is_a.name == "URL"]

    assert len(url_entities) == 2
    assert "https://www.example.com" in [e.name for e in url_entities]
    assert "http://docs.example.org/help" in [e.name for e in url_entities]


@pytest.mark.asyncio
async def test_extract_dates(regex_extractor):
    """Test extraction of dates."""
    text = "The event is scheduled for 2023-05-15 and ends on 06/30/2023."
    entities = await regex_extractor.extract_entities(text)

    # Filter only DATE entities
    date_entities = [e for e in entities if e.is_a.name == "DATE"]

    assert len(date_entities) == 2
    assert "2023-05-15" in [e.name for e in date_entities]
    assert "06/30/2023" in [e.name for e in date_entities]


@pytest.mark.asyncio
async def test_extract_times(regex_extractor):
    """Test extraction of times."""
    text = "The meeting starts at 09:30 AM and ends at 14:45."
    entities = await regex_extractor.extract_entities(text)

    # Filter only TIME entities
    time_entities = [e for e in entities if e.is_a.name == "TIME"]

    assert len(time_entities) == 2
    assert "09:30 AM" in [e.name for e in time_entities]
    assert "14:45" in [e.name for e in time_entities]


@pytest.mark.asyncio
async def test_extract_money(regex_extractor):
    """Test extraction of monetary amounts."""
    # TODO: Lazar to fix regex for test, it's failing currently
    pass
    # text = "The product costs $1,299.99 or €1.045,00 depending on your region."
    # entities = await regex_extractor.extract_entities(text)

    # Filter only MONEY entities
    # money_entities = [e for e in entities if e.is_a.name == "MONEY"]

    # assert len(money_entities) == 2
    # assert "$1,299.99" in [e.name for e in money_entities]
    # assert "€1.045,00" in [e.name for e in money_entities]


@pytest.mark.asyncio
async def test_extract_person_names(regex_extractor):
    """Test extraction of person names with various formats."""
    text = """
    Standard names: John Smith and Sarah Johnson will be attending.
    Names with titles: Dr. Jane Wilson and Prof Michael Brown will present.
    Names with middle initials: James T. Kirk and William H Gates are invited.
    Names with prefixes: Jean de la Fontaine and Ludwig van Beethoven are famous.

    Single names like Mary or Robert should not be extracted as they could be
    confused with regular capitalized words at the beginning of sentences.
    """
    entities = await regex_extractor.extract_entities(text)

    # Filter only PERSON entities
    person_entities = [e for e in entities if e.is_a.name == "PERSON"]
    entity_names = [e.name for e in person_entities]

    # Standard two-part names
    assert "John Smith" in entity_names
    assert "Sarah Johnson" in entity_names

    # Names with titles
    assert "Dr. Jane Wilson" in entity_names
    assert "Prof Michael Brown" in entity_names

    # Names with middle initials
    assert "James T. Kirk" in entity_names
    assert "William H Gates" in entity_names

    # Names with prefixes
    assert "Jean de la Fontaine" in entity_names
    assert "Ludwig van Beethoven" in entity_names

    # Verify single names are not extracted
    assert "Mary" not in entity_names
    assert "Robert" not in entity_names

    # Verify we have the expected number of names
    assert len(person_entities) == 8


@pytest.mark.asyncio
async def test_extract_hashtags(regex_extractor):
    """Test extraction of hashtags."""
    text = "Check out our latest post #Python #MachineLearning"
    entities = await regex_extractor.extract_entities(text)

    # Filter only HASHTAG entities
    hashtag_entities = [e for e in entities if e.is_a.name == "HASHTAG"]

    assert len(hashtag_entities) == 2
    assert "#Python" in [e.name for e in hashtag_entities]
    assert "#MachineLearning" in [e.name for e in hashtag_entities]


@pytest.mark.asyncio
async def test_extract_mentions(regex_extractor):
    """Test extraction of mentions."""
    text = "Thanks to @johndoe and @jane_smith for their contributions."
    entities = await regex_extractor.extract_entities(text)

    # Filter only MENTION entities
    mention_entities = [e for e in entities if e.is_a.name == "MENTION"]

    assert len(mention_entities) == 2
    assert "@johndoe" in [e.name for e in mention_entities]
    assert "@jane_smith" in [e.name for e in mention_entities]


@pytest.mark.asyncio
async def test_extract_ip_addresses(regex_extractor):
    """Test extraction of IP addresses with proper validation of octet ranges."""
    # Test with valid IP addresses
    text = "The server IPs are 192.168.1.1, 10.0.0.1, 255.255.255.255, and 0.0.0.0."
    entities = await regex_extractor.extract_entities(text)

    # Filter only IP_ADDRESS entities
    ip_entities = [e for e in entities if e.is_a.name == "IP_ADDRESS"]

    assert len(ip_entities) == 4
    assert "192.168.1.1" in [e.name for e in ip_entities]
    assert "10.0.0.1" in [e.name for e in ip_entities]
    assert "255.255.255.255" in [e.name for e in ip_entities]
    assert "0.0.0.0" in [e.name for e in ip_entities]


@pytest.mark.asyncio
async def test_invalid_ip_addresses(regex_extractor):
    """Test that invalid IP addresses are not extracted."""
    # Test with invalid IP addresses
    text = "Invalid IPs: 999.999.999.999, 256.256.256.256, 1.2.3.4.5, 01.102.103.104"
    entities = await regex_extractor.extract_entities(text)

    # Filter only IP_ADDRESS entities
    ip_entities = [e for e in entities if e.is_a.name == "IP_ADDRESS"]

    # None of these should be extracted as valid IPs
    assert len(ip_entities) == 1
    assert "999.999.999.999" not in [e.name for e in ip_entities]
    assert "256.256.256.256" not in [e.name for e in ip_entities]
    assert "1.2.3.4.5" not in [e.name for e in ip_entities]
    assert "01.102.103.104" not in [e.name for e in ip_entities]
    assert "1.102.103.104" in [e.name for e in ip_entities]


@pytest.mark.asyncio
async def test_extract_multiple_entity_types(regex_extractor):
    """Test extraction of multiple entity types from a single text."""
    text = """
    Contact John Doe at john.doe@example.com or +1-555-123-4567.
    Visit our website at https://www.example.com.
    The meeting is scheduled for 2023-05-15 at 09:30 AM.
    The project budget is $10,000.00.
    Follow us on social media with #Python and mention @pythonorg.
    Our server IP is 192.168.1.1.
    """

    entities = await regex_extractor.extract_entities(text)

    # Check that we have at least one entity of each type
    entity_types = [e.is_a.name for e in entities]

    assert "EMAIL" in entity_types
    assert "PHONE" in entity_types
    assert "URL" in entity_types
    assert "DATE" in entity_types
    assert "TIME" in entity_types
    assert "MONEY" in entity_types
    assert "PERSON" in entity_types
    assert "HASHTAG" in entity_types
    assert "MENTION" in entity_types
    assert "IP_ADDRESS" in entity_types


@pytest.mark.asyncio
async def test_empty_text(regex_extractor):
    """Test extraction with empty text."""
    entities = await regex_extractor.extract_entities("")
    assert len(entities) == 0


@pytest.mark.asyncio
async def test_none_text(regex_extractor):
    """Test extraction with None text."""
    entities = await regex_extractor.extract_entities(None)
    assert len(entities) == 0


@pytest.mark.asyncio
async def test_text_without_entities(regex_extractor):
    """Test extraction with text that doesn't contain any entities."""
    text = "This text does not contain any extractable entities."
    entities = await regex_extractor.extract_entities(text)
    assert len(entities) == 0


@pytest.mark.asyncio
async def test_custom_config_path(tmp_path):
    """Test extraction with a custom configuration path."""
    # Create a minimal test config file
    config_content = """[
        {
            "entity_name": "TEST_ENTITY",
            "entity_description": "Test entity type",
            "regex": "TEST\\\\d+",
            "description_template": "Test entity: {}"
        }
    ]"""

    config_path = tmp_path / "test_config.json"
    with open(config_path, "w") as f:
        f.write(config_content)

    # Create extractor with custom config
    extractor = RegexEntityExtractor(str(config_path))

    # Test extraction
    text = "This contains TEST123 and TEST456."
    entities = await extractor.extract_entities(text)

    assert len(entities) == 2
    assert all(e.is_a.name == "TEST_ENTITY" for e in entities)
    assert "TEST123" in [e.name for e in entities]
    assert "TEST456" in [e.name for e in entities]



from cognee.infrastructure.engine import DataPoint
from cognee.modules.engine.models.TableType import TableType
from typing import Optional


class TableRow(DataPoint):
    name: str
    is_a: Optional[TableType] = None
    description: str
    properties: str

    metadata: dict = {"index_fields": ["properties"]}



import pytest
from cognee.modules.visualization.cognee_network_visualization import (
    cognee_network_visualization,
)


@pytest.mark.asyncio
async def test_create_cognee_style_network_with_logo():
    nodes_data = [
        (1, {"type": "Entity", "name": "Node1", "updated_at": 123, "created_at": 123}),
        (
            2,
            {
                "type": "DocumentChunk",
                "name": "Node2",
                "updated_at": 123,
                "created_at": 123,
            },
        ),
    ]
    edges_data = [
        (1, 2, "related_to", {}),
    ]
    graph_data = (nodes_data, edges_data)

    html_output = await cognee_network_visualization(graph_data)

    assert isinstance(html_output, str)

    assert "<html>" in html_output
    assert '<script src="https://d3js.org/d3.v5.min.js"></script>' in html_output
    assert "var nodes =" in html_output
    assert "var links =" in html_output



import json
import uuid
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from cognee.exceptions import InvalidValueError
from cognee.modules.search.methods.search import search, specific_search
from cognee.modules.search.types import SearchType
from cognee.modules.users.models import User
import sys

search_module = sys.modules.get("cognee.modules.search.methods.search")


@pytest.fixture
def mock_user():
    user = MagicMock(spec=User)
    user.id = uuid.uuid4()
    return user


@pytest.mark.asyncio
@patch.object(search_module, "log_query")
@patch.object(search_module, "log_result")
@patch.object(search_module, "get_document_ids_for_user")
@patch.object(search_module, "specific_search")
@patch.object(search_module, "parse_id")
async def test_search(
    mock_parse_id,
    mock_specific_search,
    mock_get_document_ids,
    mock_log_result,
    mock_log_query,
    mock_user,
):
    # Setup
    query_text = "test query"
    query_type = SearchType.CHUNKS
    datasets = ["dataset1", "dataset2"]

    # Mock the query logging
    mock_query = MagicMock()
    mock_query.id = uuid.uuid4()
    mock_log_query.return_value = mock_query

    # Mock document IDs
    doc_id1 = uuid.uuid4()
    doc_id2 = uuid.uuid4()
    doc_id3 = uuid.uuid4()  # This one will be filtered out
    mock_get_document_ids.return_value = [doc_id1, doc_id2]

    # Mock search results
    search_results = [
        {"document_id": str(doc_id1), "content": "Result 1"},
        {"document_id": str(doc_id2), "content": "Result 2"},
        {"document_id": str(doc_id3), "content": "Result 3"},  # Should be filtered out
    ]
    mock_specific_search.return_value = search_results

    # Mock parse_id to return the same UUID
    mock_parse_id.side_effect = lambda x: uuid.UUID(x) if x else None

    # Execute
    results = await search(query_text, query_type, datasets, mock_user)

    # Verify
    mock_log_query.assert_called_once_with(query_text, query_type.value, mock_user.id)
    mock_get_document_ids.assert_called_once_with(mock_user.id, datasets)
    mock_specific_search.assert_called_once_with(
        query_type, query_text, mock_user, system_prompt_path="answer_simple_question.txt", top_k=10
    )

    # Only the first two results should be included (doc_id3 is filtered out)
    assert len(results) == 2
    assert results[0]["document_id"] == str(doc_id1)
    assert results[1]["document_id"] == str(doc_id2)

    # Verify result logging
    mock_log_result.assert_called_once()
    # Check that the first argument is the query ID
    assert mock_log_result.call_args[0][0] == mock_query.id
    # The second argument should be the JSON string of the filtered results
    # We can't directly compare the JSON strings due to potential ordering differences
    # So we parse the JSON and compare the objects
    logged_results = json.loads(mock_log_result.call_args[0][1])
    assert len(logged_results) == 2
    assert logged_results[0]["document_id"] == str(doc_id1)
    assert logged_results[1]["document_id"] == str(doc_id2)


@pytest.mark.asyncio
@patch.object(search_module, "SummariesRetriever")
@patch.object(search_module, "send_telemetry")
async def test_specific_search_summaries(mock_send_telemetry, mock_summaries_retriever, mock_user):
    # Setup
    query = "test query"
    query_type = SearchType.SUMMARIES

    # Mock the retriever
    mock_retriever = MagicMock()
    mock_retriever.get_completion = AsyncMock()
    mock_retriever.get_completion.return_value = [{"content": "Summary result"}]
    mock_summaries_retriever.return_value = mock_retriever

    # Execute
    results = await specific_search(query_type, query, mock_user)

    # Verify
    mock_summaries_retriever.assert_called_once()
    mock_retriever.get_completion.assert_called_once_with(query)
    mock_send_telemetry.assert_called()
    assert len(results) == 1
    assert results[0]["content"] == "Summary result"


@pytest.mark.asyncio
@patch.object(search_module, "InsightsRetriever")
@patch.object(search_module, "send_telemetry")
async def test_specific_search_insights(mock_send_telemetry, mock_insights_retriever, mock_user):
    # Setup
    query = "test query"
    query_type = SearchType.INSIGHTS

    # Mock the retriever
    mock_retriever = MagicMock()
    mock_retriever.get_completion = AsyncMock()
    mock_retriever.get_completion.return_value = [{"content": "Insight result"}]
    mock_insights_retriever.return_value = mock_retriever

    # Execute
    results = await specific_search(query_type, query, mock_user)

    # Verify
    mock_insights_retriever.assert_called_once()
    mock_retriever.get_completion.assert_called_once_with(query)
    mock_send_telemetry.assert_called()
    assert len(results) == 1
    assert results[0]["content"] == "Insight result"


@pytest.mark.asyncio
@patch.object(search_module, "ChunksRetriever")
@patch.object(search_module, "send_telemetry")
async def test_specific_search_chunks(mock_send_telemetry, mock_chunks_retriever, mock_user):
    # Setup
    query = "test query"
    query_type = SearchType.CHUNKS

    # Mock the retriever
    mock_retriever = MagicMock()
    mock_retriever.get_completion = AsyncMock()
    mock_retriever.get_completion.return_value = [{"content": "Chunk result"}]
    mock_chunks_retriever.return_value = mock_retriever

    # Execute
    results = await specific_search(query_type, query, mock_user)

    # Verify
    mock_chunks_retriever.assert_called_once()
    mock_retriever.get_completion.assert_called_once_with(query)
    mock_send_telemetry.assert_called()
    assert len(results) == 1
    assert results[0]["content"] == "Chunk result"


@pytest.mark.asyncio
async def test_specific_search_invalid_type(mock_user):
    # Setup
    query = "test query"
    query_type = "INVALID_TYPE"  # Not a valid SearchType

    # Execute and verify
    with pytest.raises(InvalidValueError) as excinfo:
        await specific_search(query_type, query, mock_user)

    assert "Unsupported search type" in str(excinfo.value)



from typing import AsyncGenerator
from fastapi import Depends
from contextlib import asynccontextmanager
from sqlalchemy.ext.asyncio import AsyncSession
from fastapi_users.db import SQLAlchemyUserDatabase
from cognee.infrastructure.databases.relational import get_relational_engine
from .models.User import User


async def get_async_session() -> AsyncGenerator[AsyncSession, None]:
    db_engine = get_relational_engine()
    async with db_engine.get_async_session() as session:
        yield session


async def get_user_db(session: AsyncSession = Depends(get_async_session)):
    yield SQLAlchemyUserDatabase(session, User)


get_user_db_context = asynccontextmanager(get_user_db)



import os
import pytest
import pathlib

import cognee
from cognee.low_level import setup
from cognee.tasks.storage import add_data_points
from cognee.infrastructure.databases.vector import get_vector_engine
from cognee.modules.chunking.models import DocumentChunk
from cognee.tasks.summarization.models import TextSummary
from cognee.modules.data.processing.document_types import TextDocument
from cognee.modules.retrieval.exceptions.exceptions import NoDataError
from cognee.modules.retrieval.summaries_retriever import SummariesRetriever


class TextSummariesRetriever:
    @pytest.mark.asyncio
    async def test_chunk_context(self):
        system_directory_path = os.path.join(
            pathlib.Path(__file__).parent, ".cognee_system/test_summary_context"
        )
        cognee.config.system_root_directory(system_directory_path)
        data_directory_path = os.path.join(
            pathlib.Path(__file__).parent, ".data_storage/test_summary_context"
        )
        cognee.config.data_root_directory(data_directory_path)

        await cognee.prune.prune_data()
        await cognee.prune.prune_system(metadata=True)
        await setup()

        document1 = TextDocument(
            name="Employee List",
            raw_data_location="somewhere",
            external_metadata="",
            mime_type="text/plain",
        )

        document2 = TextDocument(
            name="Car List",
            raw_data_location="somewhere",
            external_metadata="",
            mime_type="text/plain",
        )

        chunk1 = DocumentChunk(
            text="Steve Rodger",
            chunk_size=2,
            chunk_index=0,
            cut_type="sentence_end",
            is_part_of=document1,
            contains=[],
        )
        chunk1_summary = TextSummary(
            text="S.R.",
            made_from=chunk1,
        )
        chunk2 = DocumentChunk(
            text="Mike Broski",
            chunk_size=2,
            chunk_index=1,
            cut_type="sentence_end",
            is_part_of=document1,
            contains=[],
        )
        chunk2_summary = TextSummary(
            text="M.B.",
            made_from=chunk2,
        )
        chunk3 = DocumentChunk(
            text="Christina Mayer",
            chunk_size=2,
            chunk_index=2,
            cut_type="sentence_end",
            is_part_of=document1,
            contains=[],
        )
        chunk3_summary = TextSummary(
            text="C.M.",
            made_from=chunk3,
        )
        chunk4 = DocumentChunk(
            text="Range Rover",
            chunk_size=2,
            chunk_index=0,
            cut_type="sentence_end",
            is_part_of=document2,
            contains=[],
        )
        chunk4_summary = TextSummary(
            text="R.R.",
            made_from=chunk4,
        )
        chunk5 = DocumentChunk(
            text="Hyundai",
            chunk_size=2,
            chunk_index=1,
            cut_type="sentence_end",
            is_part_of=document2,
            contains=[],
        )
        chunk5_summary = TextSummary(
            text="H.Y.",
            made_from=chunk5,
        )
        chunk6 = DocumentChunk(
            text="Chrysler",
            chunk_size=2,
            chunk_index=2,
            cut_type="sentence_end",
            is_part_of=document2,
            contains=[],
        )
        chunk6_summary = TextSummary(
            text="C.H.",
            made_from=chunk6,
        )

        entities = [
            chunk1_summary,
            chunk2_summary,
            chunk3_summary,
            chunk4_summary,
            chunk5_summary,
            chunk6_summary,
        ]

        await add_data_points(entities)

        retriever = SummariesRetriever(limit=20)

        context = await retriever.get_context("Christina")

        assert context[0]["text"] == "C.M.", "Failed to get Christina Mayer"

    @pytest.mark.asyncio
    async def test_chunk_context_on_empty_graph(self):
        system_directory_path = os.path.join(
            pathlib.Path(__file__).parent, ".cognee_system/test_summary_context"
        )
        cognee.config.system_root_directory(system_directory_path)
        data_directory_path = os.path.join(
            pathlib.Path(__file__).parent, ".data_storage/test_summary_context"
        )
        cognee.config.data_root_directory(data_directory_path)

        await cognee.prune.prune_data()
        await cognee.prune.prune_system(metadata=True)

        retriever = SummariesRetriever()

        with pytest.raises(NoDataError):
            await retriever.get_context("Christina Mayer")

        vector_engine = get_vector_engine()
        await vector_engine.create_collection("TextSummary_text", payload_schema=TextSummary)

        context = await retriever.get_context("Christina Mayer")
        assert context == [], "Returned context should be empty on an empty graph"


if __name__ == "__main__":
    from asyncio import run

    test = TextSummariesRetriever()

    run(test.test_chunk_context())
    run(test.test_chunk_context_on_empty_graph())



from typing import List
from cognee.api.v1.cognify.cognify import get_default_tasks
from cognee.modules.pipelines.tasks.task import Task
from cognee.modules.chunking.TextChunker import TextChunker
from cognee.tasks.graph import extract_graph_from_data
from cognee.tasks.storage import add_data_points
from cognee.shared.data_models import KnowledgeGraph
from cognee.modules.ontology.rdf_xml.OntologyResolver import OntologyResolver


async def get_default_tasks_by_indices(
    indices: List[int], chunk_size: int = None, chunker=TextChunker
) -> List[Task]:
    """Returns default tasks filtered by the provided indices."""
    all_tasks = await get_default_tasks(chunker=chunker, chunk_size=chunk_size)

    if any(i < 0 or i >= len(all_tasks) for i in indices):
        raise IndexError(
            f"Task indices {indices} out of range. Valid range: 0-{len(all_tasks) - 1}"
        )

    return [all_tasks[i] for i in indices]


async def get_no_summary_tasks(
    chunk_size: int = None,
    chunker=TextChunker,
    user=None,
    graph_model=KnowledgeGraph,
    ontology_file_path=None,
) -> List[Task]:
    """Returns default tasks without summarization tasks."""
    # Get base tasks (0=classify, 1=check_permissions, 2=extract_chunks)
    base_tasks = await get_default_tasks_by_indices([0, 1, 2], chunk_size, chunker)

    ontology_adapter = OntologyResolver(ontology_file=ontology_file_path)

    graph_task = Task(
        extract_graph_from_data,
        graph_model=graph_model,
        ontology_adapter=ontology_adapter,
        task_config={"batch_size": 10},
    )

    add_data_points_task = Task(add_data_points, task_config={"batch_size": 10})

    return base_tasks + [graph_task, add_data_points_task]


async def get_just_chunks_tasks(
    chunk_size: int = None, chunker=TextChunker, user=None
) -> List[Task]:
    """Returns default tasks with only chunk extraction and data points addition."""
    # Get base tasks (0=classify, 1=check_permissions, 2=extract_chunks)
    base_tasks = await get_default_tasks_by_indices([0, 1, 2], chunk_size, chunker)

    add_data_points_task = Task(add_data_points, task_config={"batch_size": 10})

    return base_tasks + [add_data_points_task]



import os
from functools import lru_cache

import dlt
from typing import Union

from cognee.infrastructure.databases.relational import get_relational_config


@lru_cache
def get_dlt_destination() -> Union[type[dlt.destinations.sqlalchemy], None]:
    """
    Handles propagation of the cognee database configuration to the dlt library

    Returns:
        sqlachemy: sqlachemy destination used by the dlt library

    """
    relational_config = get_relational_config()

    if relational_config.db_provider == "sqlite":
        # When sqlite is the database provider hostname, port, username and password should not be forwarded.
        # The database is found  by combining the path location and the database name
        destination = dlt.destinations.sqlalchemy(
            credentials={
                "database": os.path.join(relational_config.db_path, relational_config.db_name),
                "drivername": relational_config.db_provider,
            },
        )
    elif relational_config.db_provider == "postgres":
        # The dlt library doesn't accept postgres as the drivername, it only accepts postgresql
        destination = dlt.destinations.sqlalchemy(
            credentials={
                "host": relational_config.db_host,
                "port": relational_config.db_port,
                "username": relational_config.db_username,
                "password": relational_config.db_password,
                "database": relational_config.db_name,
                "drivername": "postgresql",
            },
        )
    else:
        destination = None

    return destination



import os
import pytest
import networkx as nx
import pandas as pd
from unittest.mock import patch, mock_open
from io import BytesIO
from uuid import uuid4

from cognee.shared.utils import (
    get_anonymous_id,
    get_file_content_hash,
    prepare_edges,
    prepare_nodes,
)


@pytest.fixture
def temp_dir(tmp_path):
    return tmp_path


@patch("os.makedirs")
@patch("builtins.open", new_callable=mock_open, read_data=str(uuid4()))
def test_get_anonymous_id(mock_open_file, mock_makedirs, temp_dir):
    os.environ["HOME"] = str(temp_dir)
    anon_id = get_anonymous_id()
    assert isinstance(anon_id, str)
    assert len(anon_id) > 0


# @patch("requests.post")
# def test_send_telemetry(mock_post):
#     mock_post.return_value.status_code = 200
#
#     send_telemetry("test_event", "test_user", {"key": "value"})
#     mock_post.assert_called_once()
#
#     args, kwargs = mock_post.call_args
#     assert kwargs["json"]["event_name"] == "test_event"


@patch("builtins.open", new_callable=mock_open, read_data=b"test_data")
def test_get_file_content_hash_file(mock_open_file):
    import hashlib

    expected_hash = hashlib.md5(b"test_data").hexdigest()
    result = get_file_content_hash("test_file.txt")
    assert result == expected_hash


def test_get_file_content_hash_stream():
    stream = BytesIO(b"test_data")
    import hashlib

    expected_hash = hashlib.md5(b"test_data").hexdigest()
    result = get_file_content_hash(stream)
    assert result == expected_hash


def test_prepare_edges():
    graph = nx.MultiDiGraph()
    graph.add_edge("A", "B", key="AB", weight=1)
    edges_df = prepare_edges(graph, "source", "target", "key")

    assert isinstance(edges_df, pd.DataFrame)
    assert len(edges_df) == 1


def test_prepare_nodes():
    graph = nx.Graph()
    graph.add_node(1, name="Node1")
    nodes_df = prepare_nodes(graph)

    assert isinstance(nodes_df, pd.DataFrame)
    assert len(nodes_df) == 1



#

"""Neo4j Adapter for Graph Database"""

import json
from cognee.shared.logging_utils import get_logger, ERROR
import asyncio
from textwrap import dedent
from typing import Optional, Any, List, Dict
from contextlib import asynccontextmanager
from uuid import UUID
from neo4j import AsyncSession
from neo4j import AsyncGraphDatabase
from neo4j.exceptions import Neo4jError
from cognee.infrastructure.engine import DataPoint
from cognee.infrastructure.databases.graph.graph_db_interface import (
    GraphDBInterface,
    record_graph_changes,
)
from cognee.modules.storage.utils import JSONEncoder
from .neo4j_metrics_utils import (
    get_avg_clustering,
    get_edge_density,
    get_num_connected_components,
    get_shortest_path_lengths,
    get_size_of_connected_components,
    count_self_loops,
)

logger = get_logger("Neo4jAdapter", level=ERROR)


class Neo4jAdapter(GraphDBInterface):
    def __init__(
        self,
        graph_database_url: str,
        graph_database_username: str,
        graph_database_password: str,
        driver: Optional[Any] = None,
    ):
        self.driver = driver or AsyncGraphDatabase.driver(
            graph_database_url,
            auth=(graph_database_username, graph_database_password),
            max_connection_lifetime=120,
        )

    @asynccontextmanager
    async def get_session(self) -> AsyncSession:
        async with self.driver.session() as session:
            yield session

    async def query(
        self,
        query: str,
        params: Optional[Dict[str, Any]] = None,
    ) -> List[Dict[str, Any]]:
        try:
            async with self.get_session() as session:
                result = await session.run(query, parameters=params)
                data = await result.data()
                return data
        except Neo4jError as error:
            logger.error("Neo4j query error: %s", error, exc_info=True)
            raise error

    async def has_node(self, node_id: str) -> bool:
        results = self.query(
            """
                MATCH (n)
                WHERE n.id = $node_id
                RETURN COUNT(n) > 0 AS node_exists
            """,
            {"node_id": node_id},
        )
        return results[0]["node_exists"] if len(results) > 0 else False

    async def add_node(self, node: DataPoint):
        serialized_properties = self.serialize_properties(node.model_dump())

        query = dedent(
            """MERGE (node {id: $node_id})
                ON CREATE SET node += $properties, node.updated_at = timestamp()
                ON MATCH SET node += $properties, node.updated_at = timestamp()
                WITH node, $node_label AS label
                CALL apoc.create.addLabels(node, [label]) YIELD node AS labeledNode
                RETURN ID(labeledNode) AS internal_id, labeledNode.id AS nodeId"""
        )

        params = {
            "node_id": str(node.id),
            "node_label": type(node).__name__,
            "properties": serialized_properties,
        }

        return await self.query(query, params)

    @record_graph_changes
    async def add_nodes(self, nodes: list[DataPoint]) -> None:
        query = """
        UNWIND $nodes AS node
        MERGE (n {id: node.node_id})
        ON CREATE SET n += node.properties, n.updated_at = timestamp()
        ON MATCH SET n += node.properties, n.updated_at = timestamp()
        WITH n, node.label AS label
        CALL apoc.create.addLabels(n, [label]) YIELD node AS labeledNode
        RETURN ID(labeledNode) AS internal_id, labeledNode.id AS nodeId
        """

        nodes = [
            {
                "node_id": str(node.id),
                "label": type(node).__name__,
                "properties": self.serialize_properties(node.model_dump()),
            }
            for node in nodes
        ]

        results = await self.query(query, dict(nodes=nodes))
        return results

    async def extract_node(self, node_id: str):
        results = await self.extract_nodes([node_id])

        return results[0] if len(results) > 0 else None

    async def extract_nodes(self, node_ids: List[str]):
        query = """
        UNWIND $node_ids AS id
        MATCH (node {id: id})
        RETURN node"""

        params = {"node_ids": node_ids}

        results = await self.query(query, params)

        return [result["node"] for result in results]

    async def delete_node(self, node_id: str):
        query = "MATCH (node {id: $node_id}) DETACH DELETE node"
        params = {"node_id": node_id}

        return await self.query(query, params)

    async def delete_nodes(self, node_ids: list[str]) -> None:
        query = """
        UNWIND $node_ids AS id
        MATCH (node {id: id})
        DETACH DELETE node"""

        params = {"node_ids": node_ids}

        return await self.query(query, params)

    async def has_edge(self, from_node: UUID, to_node: UUID, edge_label: str) -> bool:
        query = """
            MATCH (from_node)-[relationship]->(to_node)
            WHERE from_node.id = $from_node_id AND to_node.id = $to_node_id AND type(relationship) = $edge_label
            RETURN COUNT(relationship) > 0 AS edge_exists
        """

        params = {
            "from_node_id": str(from_node),
            "to_node_id": str(to_node),
            "edge_label": edge_label,
        }

        edge_exists = await self.query(query, params)
        return edge_exists

    async def has_edges(self, edges):
        query = """
            UNWIND $edges AS edge
            MATCH (a)-[r]->(b)
            WHERE id(a) = edge.from_node AND id(b) = edge.to_node AND type(r) = edge.relationship_name
            RETURN edge.from_node AS from_node, edge.to_node AS to_node, edge.relationship_name AS relationship_name, count(r) > 0 AS edge_exists
        """

        try:
            params = {
                "edges": [
                    {
                        "from_node": str(edge[0]),
                        "to_node": str(edge[1]),
                        "relationship_name": edge[2],
                    }
                    for edge in edges
                ],
            }

            results = await self.query(query, params)
            return [result["edge_exists"] for result in results]
        except Neo4jError as error:
            logger.error("Neo4j query error: %s", error, exc_info=True)
            raise error

    async def add_edge(
        self,
        from_node: UUID,
        to_node: UUID,
        relationship_name: str,
        edge_properties: Optional[Dict[str, Any]] = {},
    ):
        serialized_properties = self.serialize_properties(edge_properties)

        query = dedent(
            f"""\
            MATCH (from_node {{id: $from_node}}),
                  (to_node {{id: $to_node}})
            MERGE (from_node)-[r:{relationship_name}]->(to_node)
            ON CREATE SET r += $properties, r.updated_at = timestamp()
            ON MATCH SET r += $properties, r.updated_at = timestamp()
            RETURN r
            """
        )

        params = {
            "from_node": str(from_node),
            "to_node": str(to_node),
            "relationship_name": relationship_name,
            "properties": serialized_properties,
        }

        return await self.query(query, params)

    @record_graph_changes
    async def add_edges(self, edges: list[tuple[str, str, str, dict[str, Any]]]) -> None:
        query = """
            UNWIND $edges AS edge
            MATCH (from_node {id: edge.from_node})
            MATCH (to_node {id: edge.to_node})
            CALL apoc.merge.relationship(
                from_node,
                edge.relationship_name,
                {
                    source_node_id: edge.from_node,
                    target_node_id: edge.to_node
                },
                edge.properties,
                to_node
            ) YIELD rel
            RETURN rel"""

        edges = [
            {
                "from_node": str(edge[0]),
                "to_node": str(edge[1]),
                "relationship_name": edge[2],
                "properties": {
                    **(edge[3] if edge[3] else {}),
                    "source_node_id": str(edge[0]),
                    "target_node_id": str(edge[1]),
                },
            }
            for edge in edges
        ]

        try:
            results = await self.query(query, dict(edges=edges))
            return results
        except Neo4jError as error:
            logger.error("Neo4j query error: %s", error, exc_info=True)
            raise error

    async def get_edges(self, node_id: str):
        query = """
        MATCH (n {id: $node_id})-[r]-(m)
        RETURN n, r, m
        """

        results = await self.query(query, dict(node_id=node_id))

        return [
            (result["n"]["id"], result["m"]["id"], {"relationship_name": result["r"][1]})
            for result in results
        ]

    async def get_disconnected_nodes(self) -> list[str]:
        # return await self.query(
        #     "MATCH (node) WHERE NOT (node)<-[:*]-() RETURN node.id as id",
        # )
        query = """
        // Step 1: Collect all nodes
        MATCH (n)
        WITH COLLECT(n) AS nodes

        // Step 2: Find all connected components
        WITH nodes
        CALL {
          WITH nodes
          UNWIND nodes AS startNode
          MATCH path = (startNode)-[*]-(connectedNode)
          WITH startNode, COLLECT(DISTINCT connectedNode) AS component
          RETURN component
        }

        // Step 3: Aggregate components
        WITH COLLECT(component) AS components

        // Step 4: Identify the largest connected component
        UNWIND components AS component
        WITH component
        ORDER BY SIZE(component) DESC
        LIMIT 1
        WITH component AS largestComponent

        // Step 5: Find nodes not in the largest connected component
        MATCH (n)
        WHERE NOT n IN largestComponent
        RETURN COLLECT(ID(n)) AS ids
        """

        results = await self.query(query)
        return results[0]["ids"] if len(results) > 0 else []

    async def get_predecessors(self, node_id: str, edge_label: str = None) -> list[str]:
        if edge_label is not None:
            query = """
            MATCH (node)<-[r]-(predecessor)
            WHERE node.id = $node_id AND type(r) = $edge_label
            RETURN predecessor
            """

            results = await self.query(
                query,
                dict(
                    node_id=node_id,
                    edge_label=edge_label,
                ),
            )

            return [result["predecessor"] for result in results]
        else:
            query = """
            MATCH (node)<-[r]-(predecessor)
            WHERE node.id = $node_id
            RETURN predecessor
            """

            results = await self.query(
                query,
                dict(
                    node_id=node_id,
                ),
            )

            return [result["predecessor"] for result in results]

    async def get_successors(self, node_id: str, edge_label: str = None) -> list[str]:
        if edge_label is not None:
            query = """
            MATCH (node)-[r]->(successor)
            WHERE node.id = $node_id AND type(r) = $edge_label
            RETURN successor
            """

            results = await self.query(
                query,
                dict(
                    node_id=node_id,
                    edge_label=edge_label,
                ),
            )

            return [result["successor"] for result in results]
        else:
            query = """
            MATCH (node)-[r]->(successor)
            WHERE node.id = $node_id
            RETURN successor
            """

            results = await self.query(
                query,
                dict(
                    node_id=node_id,
                ),
            )

            return [result["successor"] for result in results]

    async def get_neighbors(self, node_id: str) -> List[Dict[str, Any]]:
        """Get all neighboring nodes."""
        return await self.get_neighbours(node_id)

    async def get_node(self, node_id: str) -> Optional[Dict[str, Any]]:
        """Get a single node by ID."""
        query = """
        MATCH (node {id: $node_id})
        RETURN node
        """
        results = await self.query(query, {"node_id": node_id})
        return results[0]["node"] if results else None

    async def get_nodes(self, node_ids: List[str]) -> List[Dict[str, Any]]:
        """Get multiple nodes by their IDs."""
        query = """
        UNWIND $node_ids AS id
        MATCH (node {id: id})
        RETURN node
        """
        results = await self.query(query, {"node_ids": node_ids})
        return [result["node"] for result in results]

    async def get_connections(self, node_id: UUID) -> list:
        predecessors_query = """
        MATCH (node)<-[relation]-(neighbour)
        WHERE node.id = $node_id
        RETURN neighbour, relation, node
        """
        successors_query = """
        MATCH (node)-[relation]->(neighbour)
        WHERE node.id = $node_id
        RETURN node, relation, neighbour
        """

        predecessors, successors = await asyncio.gather(
            self.query(predecessors_query, dict(node_id=str(node_id))),
            self.query(successors_query, dict(node_id=str(node_id))),
        )

        connections = []

        for neighbour in predecessors:
            neighbour = neighbour["relation"]
            connections.append((neighbour[0], {"relationship_name": neighbour[1]}, neighbour[2]))

        for neighbour in successors:
            neighbour = neighbour["relation"]
            connections.append((neighbour[0], {"relationship_name": neighbour[1]}, neighbour[2]))

        return connections

    async def remove_connection_to_predecessors_of(
        self, node_ids: list[str], edge_label: str
    ) -> None:
        query = f"""
        UNWIND $node_ids AS id
        MATCH (node:`{id}`)-[r:{edge_label}]->(predecessor)
        DELETE r;
        """

        params = {"node_ids": node_ids}

        return await self.query(query, params)

    async def remove_connection_to_successors_of(
        self, node_ids: list[str], edge_label: str
    ) -> None:
        query = f"""
        UNWIND $node_ids AS id
        MATCH (node:`{id}`)<-[r:{edge_label}]-(successor)
        DELETE r;
        """

        params = {"node_ids": node_ids}

        return await self.query(query, params)

    async def delete_graph(self):
        query = """MATCH (node)
                DETACH DELETE node;"""

        return await self.query(query)

    def serialize_properties(self, properties=dict()):
        serialized_properties = {}

        for property_key, property_value in properties.items():
            if isinstance(property_value, UUID):
                serialized_properties[property_key] = str(property_value)
                continue

            if isinstance(property_value, dict):
                serialized_properties[property_key] = json.dumps(property_value, cls=JSONEncoder)
                continue

            serialized_properties[property_key] = property_value

        return serialized_properties

    async def get_model_independent_graph_data(self):
        query_nodes = "MATCH (n) RETURN collect(n) AS nodes"
        nodes = await self.query(query_nodes)

        query_edges = "MATCH (n)-[r]->(m) RETURN collect([n, r, m]) AS elements"
        edges = await self.query(query_edges)

        return (nodes, edges)

    async def get_graph_data(self):
        query = "MATCH (n) RETURN ID(n) AS id, labels(n) AS labels, properties(n) AS properties"

        result = await self.query(query)

        nodes = [
            (
                record["properties"]["id"],
                record["properties"],
            )
            for record in result
        ]

        query = """
        MATCH (n)-[r]->(m)
        RETURN ID(n) AS source, ID(m) AS target, TYPE(r) AS type, properties(r) AS properties
        """
        result = await self.query(query)
        edges = [
            (
                record["properties"]["source_node_id"],
                record["properties"]["target_node_id"],
                record["type"],
                record["properties"],
            )
            for record in result
        ]

        return (nodes, edges)

    async def get_filtered_graph_data(self, attribute_filters):
        """
        Fetches nodes and relationships filtered by specified attribute values.

        Args:
            attribute_filters (list of dict): A list of dictionaries where keys are attributes and values are lists of values to filter on.
                                              Example: [{"community": ["1", "2"]}]

        Returns:
            tuple: A tuple containing two lists: nodes and edges.
        """
        where_clauses = []
        for attribute, values in attribute_filters[0].items():
            values_str = ", ".join(
                f"'{value}'" if isinstance(value, str) else str(value) for value in values
            )
            where_clauses.append(f"n.{attribute} IN [{values_str}]")

        where_clause = " AND ".join(where_clauses)

        query_nodes = f"""
        MATCH (n)
        WHERE {where_clause}
        RETURN ID(n) AS id, labels(n) AS labels, properties(n) AS properties
        """
        result_nodes = await self.query(query_nodes)

        nodes = [
            (
                record["id"],
                record["properties"],
            )
            for record in result_nodes
        ]

        query_edges = f"""
        MATCH (n)-[r]->(m)
        WHERE {where_clause} AND {where_clause.replace("n.", "m.")}
        RETURN ID(n) AS source, ID(m) AS target, TYPE(r) AS type, properties(r) AS properties
        """
        result_edges = await self.query(query_edges)

        edges = [
            (
                record["source"],
                record["target"],
                record["type"],
                record["properties"],
            )
            for record in result_edges
        ]

        return (nodes, edges)

    async def graph_exists(self, graph_name="myGraph"):
        query = "CALL gds.graph.list() YIELD graphName RETURN collect(graphName) AS graphNames;"
        result = await self.query(query)
        graph_names = result[0]["graphNames"] if result else []
        return graph_name in graph_names

    async def get_node_labels_string(self):
        node_labels_query = "CALL db.labels() YIELD label RETURN collect(label) AS labels;"
        node_labels_result = await self.query(node_labels_query)
        node_labels = node_labels_result[0]["labels"] if node_labels_result else []

        if not node_labels:
            raise ValueError("No node labels found in the database")

        node_labels_str = "[" + ", ".join(f"'{label}'" for label in node_labels) + "]"
        return node_labels_str

    async def get_relationship_labels_string(self):
        relationship_types_query = "CALL db.relationshipTypes() YIELD relationshipType RETURN collect(relationshipType) AS relationships;"
        relationship_types_result = await self.query(relationship_types_query)
        relationship_types = (
            relationship_types_result[0]["relationships"] if relationship_types_result else []
        )

        if not relationship_types:
            raise ValueError("No relationship types found in the database.")

        relationship_types_undirected_str = (
            "{"
            + ", ".join(f"{rel}" + ": {orientation: 'UNDIRECTED'}" for rel in relationship_types)
            + "}"
        )
        return relationship_types_undirected_str

    async def project_entire_graph(self, graph_name="myGraph"):
        """
        Projects all node labels and all relationship types into an undirected in-memory GDS graph.
        """
        if await self.graph_exists(graph_name):
            return

        node_labels_str = await self.get_node_labels_string()
        relationship_types_undirected_str = await self.get_relationship_labels_string()

        query = f"""
        CALL gds.graph.project(
            '{graph_name}',
            {node_labels_str},
            {relationship_types_undirected_str}
        ) YIELD graphName;
        """

        await self.query(query)

    async def drop_graph(self, graph_name="myGraph"):
        if await self.graph_exists(graph_name):
            drop_query = f"CALL gds.graph.drop('{graph_name}');"
            await self.query(drop_query)

    async def get_graph_metrics(self, include_optional=False):
        """For the definition of these metrics, please refer to
        https://docs.cognee.ai/core_concepts/graph_generation/descriptive_metrics"""

        nodes, edges = await self.get_model_independent_graph_data()
        graph_name = "myGraph"
        await self.drop_graph(graph_name)
        await self.project_entire_graph(graph_name)

        num_nodes = len(nodes[0]["nodes"])
        num_edges = len(edges[0]["elements"])

        mandatory_metrics = {
            "num_nodes": num_nodes,
            "num_edges": num_edges,
            "mean_degree": (2 * num_edges) / num_nodes if num_nodes != 0 else None,
            "edge_density": await get_edge_density(self),
            "num_connected_components": await get_num_connected_components(self, graph_name),
            "sizes_of_connected_components": await get_size_of_connected_components(
                self, graph_name
            ),
        }

        if include_optional:
            shortest_path_lengths = await get_shortest_path_lengths(self, graph_name)
            optional_metrics = {
                "num_selfloops": await count_self_loops(self),
                "diameter": max(shortest_path_lengths) if shortest_path_lengths else -1,
                "avg_shortest_path_length": sum(shortest_path_lengths) / len(shortest_path_lengths)
                if shortest_path_lengths
                else -1,
                "avg_clustering": await get_avg_clustering(self, graph_name),
            }
        else:
            optional_metrics = {
                "num_selfloops": -1,
                "diameter": -1,
                "avg_shortest_path_length": -1,
                "avg_clustering": -1,
            }

        return mandatory_metrics | optional_metrics

    async def get_document_subgraph(self, content_hash: str):
        query = """
        MATCH (doc)
        WHERE (doc:TextDocument OR doc:PdfDocument)
        AND doc.name = 'text_' + $content_hash

        OPTIONAL MATCH (doc)<-[:is_part_of]-(chunk:DocumentChunk)
        OPTIONAL MATCH (chunk)-[:contains]->(entity:Entity)
        WHERE NOT EXISTS {
            MATCH (entity)<-[:contains]-(otherChunk:DocumentChunk)-[:is_part_of]->(otherDoc)
            WHERE (otherDoc:TextDocument OR otherDoc:PdfDocument)
            AND otherDoc.id <> doc.id
        }
        OPTIONAL MATCH (chunk)<-[:made_from]-(made_node:TextSummary)
        OPTIONAL MATCH (entity)-[:is_a]->(type:EntityType)
        WHERE NOT EXISTS {
            MATCH (type)<-[:is_a]-(otherEntity:Entity)<-[:contains]-(otherChunk:DocumentChunk)-[:is_part_of]->(otherDoc)
            WHERE (otherDoc:TextDocument OR otherDoc:PdfDocument)
            AND otherDoc.id <> doc.id
        }

        RETURN
            collect(DISTINCT doc) as document,
            collect(DISTINCT chunk) as chunks,
            collect(DISTINCT entity) as orphan_entities,
            collect(DISTINCT made_node) as made_from_nodes,
            collect(DISTINCT type) as orphan_types
        """
        result = await self.query(query, {"content_hash": content_hash})
        return result[0] if result else None

    async def get_degree_one_nodes(self, node_type: str):
        if not node_type or node_type not in ["Entity", "EntityType"]:
            raise ValueError("node_type must be either 'Entity' or 'EntityType'")

        query = f"""
        MATCH (n:{node_type})
        WHERE COUNT {{ MATCH (n)--() }} = 1
        RETURN n
        """
        result = await self.query(query)
        return [record["n"] for record in result] if result else []



from cognee.modules.data.processing.document_types import Document
from cognee.modules.users.permissions.methods import check_permission_on_documents
from typing import List


async def check_permissions_on_documents(
    documents: list[Document], user, permissions
) -> List[Document]:
    """
    Validates a user's permissions on a list of documents.

    Notes:
        - This function assumes that `check_permission_on_documents` raises an exception if the permission check fails.
        - It is designed to validate multiple permissions in a sequential manner for the same set of documents.
        - Ensure that the `Document` and `user` objects conform to the expected structure and interfaces.
    """
    document_ids = [document.id for document in documents]

    for permission in permissions:
        await check_permission_on_documents(
            user,
            permission,
            document_ids,
        )

    return documents



from collections import Counter
from deepeval.test_case import LLMTestCase
import re
from typing import Optional, Any


class F1ScoreMetric:
    def __init__(self) -> None:
        self.score: Optional[float] = None
        self.reason: Optional[str] = None

    def measure(self, test_case: "LLMTestCase") -> float:
        actual = (test_case.actual_output or "").lower()
        expected = (test_case.expected_output or "").lower()

        actual_tokens = [
            re.sub(r"\W+", "", token.strip())
            for token in actual.split()
            if re.sub(r"\W+", "", token.strip())
        ]

        expected_tokens = [
            re.sub(r"\W+", "", token.strip())
            for token in expected.split()
            if re.sub(r"\W+", "", token.strip())
        ]

        if not actual_tokens and not expected_tokens:
            self.score = 1.0
            self.reason = "Both actual and expected are empty"
            return self.score

        actual_counts = Counter(actual_tokens)
        expected_counts = Counter(expected_tokens)

        tp = sum(min(actual_counts[word], expected_counts[word]) for word in actual_counts)
        fp = sum(actual_counts[word] for word in actual_counts) - tp
        fn = sum(expected_counts[word] for word in expected_counts) - tp

        precision = tp / (tp + fp) if (tp + fp) > 0 else 0.0
        recall = tp / (tp + fn) if (tp + fn) > 0 else 0.0

        f1 = 2 * (precision * recall) / (precision + recall) if (precision + recall) > 0 else 0.0
        self.score = f1
        self.reason = f"F1: {f1:.2f} (Precision: {precision:.2f}, Recall: {recall:.2f})"
        return self.score



from cognee.shared.logging_utils import get_logger
from cognee.modules.users.methods import get_default_user
from uuid import UUID
from sqlalchemy import select
from sqlalchemy.orm import joinedload

from cognee.modules.users.exceptions import PermissionDeniedError
from cognee.infrastructure.databases.relational import get_relational_engine

from ...models.User import User
from ...models.ACL import ACL

logger = get_logger()


async def check_permission_on_documents(user: User, permission_type: str, document_ids: list[UUID]):
    if user is None:
        user = await get_default_user()

    # TODO: Enable user role permissions again. Temporarily disabled during rework.
    # user_roles_ids = [role.id for role in user.roles]
    user_roles_ids = []

    db_engine = get_relational_engine()

    async with db_engine.get_async_session() as session:
        result = await session.execute(
            select(ACL)
            .join(ACL.permission)
            .options(joinedload(ACL.data))
            .where(ACL.principal_id.in_([user.id, *user_roles_ids]))
            .where(ACL.permission.has(name=permission_type))
        )
        acls = result.unique().scalars().all()
        data_ids = [acl.data.id for acl in acls]
        has_permissions = all(document_id in data_ids for document_id in document_ids)

        if not has_permissions:
            raise PermissionDeniedError(
                message=f"User {user.id} does not have {permission_type} permission on documents"
            )



from typing import Optional, Any, List, Union, Tuple

from cognee.eval_framework.benchmark_adapters.base_benchmark_adapter import BaseBenchmarkAdapter


class DummyAdapter(BaseBenchmarkAdapter):
    def load_corpus(
        self,
        limit: Optional[int] = None,
        seed: int = 42,
        load_golden_context: bool = False,
        instance_filter: Optional[Union[str, List[str], List[int]]] = None,
    ) -> Tuple[List[str], List[dict[str, Any]]]:
        corpus_list = [
            "The cognee is an AI memory engine that supports different vector and graph databases",
            "Neo4j is a graph database supported by cognee",
        ]
        qa_pair = {
            "answer": "Yes",
            "question": "Is Neo4j supported by cognee?",
            "type": "dummy",
        }

        if load_golden_context:
            qa_pair["golden_context"] = "Cognee supports Neo4j and NetworkX"

        question_answer_pairs = [qa_pair]

        # Instance filtering is not applicable for the dummy adapter as it always returns the same data
        # but we include the parameter for API consistency

        return corpus_list, question_answer_pairs



import asyncio
from cognee.shared.logging_utils import get_logger, ERROR
from typing import List, Optional

from cognee.infrastructure.databases.graph import get_graph_engine
from cognee.infrastructure.databases.vector import get_vector_engine
from cognee.modules.graph.cognee_graph.CogneeGraph import CogneeGraph
from cognee.modules.users.methods import get_default_user
from cognee.modules.users.models import User
from cognee.shared.utils import send_telemetry
from cognee.modules.retrieval.exceptions import CollectionDistancesNotFoundError

logger = get_logger(level=ERROR)


def format_triplets(edges):
    print("\n\n\n")

    def filter_attributes(obj, attributes):
        """Helper function to filter out non-None properties, including nested dicts."""
        result = {}
        for attr in attributes:
            value = getattr(obj, attr, None)
            if value is not None:
                # If the value is a dict, extract relevant keys from it
                if isinstance(value, dict):
                    nested_values = {
                        k: v for k, v in value.items() if k in attributes and v is not None
                    }
                    result[attr] = nested_values
                else:
                    result[attr] = value
        return result

    triplets = []
    for edge in edges:
        node1 = edge.node1
        node2 = edge.node2
        edge_attributes = edge.attributes
        node1_attributes = node1.attributes
        node2_attributes = node2.attributes

        # Filter only non-None properties
        node1_info = {key: value for key, value in node1_attributes.items() if value is not None}
        node2_info = {key: value for key, value in node2_attributes.items() if value is not None}
        edge_info = {key: value for key, value in edge_attributes.items() if value is not None}

        # Create the formatted triplet
        triplet = f"Node1: {node1_info}\nEdge: {edge_info}\nNode2: {node2_info}\n\n\n"
        triplets.append(triplet)

    return "".join(triplets)


async def get_memory_fragment(
    properties_to_project: Optional[List[str]] = None,
) -> CogneeGraph:
    """Creates and initializes a CogneeGraph memory fragment with optional property projections."""
    graph_engine = await get_graph_engine()
    memory_fragment = CogneeGraph()

    if properties_to_project is None:
        properties_to_project = ["id", "description", "name", "type", "text"]

    await memory_fragment.project_graph_from_db(
        graph_engine,
        node_properties_to_project=properties_to_project,
        edge_properties_to_project=["relationship_name"],
    )

    return memory_fragment


async def brute_force_triplet_search(
    query: str,
    user: User = None,
    top_k: int = 5,
    collections: List[str] = None,
    properties_to_project: List[str] = None,
    memory_fragment: Optional[CogneeGraph] = None,
) -> list:
    if user is None:
        user = await get_default_user()

    retrieved_results = await brute_force_search(
        query,
        user,
        top_k,
        collections=collections,
        properties_to_project=properties_to_project,
        memory_fragment=memory_fragment,
    )
    return retrieved_results


async def brute_force_search(
    query: str,
    user: User,
    top_k: int,
    collections: List[str] = None,
    properties_to_project: List[str] = None,
    memory_fragment: Optional[CogneeGraph] = None,
) -> list:
    """
    Performs a brute force search to retrieve the top triplets from the graph.

    Args:
        query (str): The search query.
        user (User): The user performing the search.
        top_k (int): The number of top results to retrieve.
        collections (Optional[List[str]]): List of collections to query.
        properties_to_project (Optional[List[str]]): List of properties to project.
        memory_fragment (Optional[CogneeGraph]): Existing memory fragment to reuse.

    Returns:
        list: The top triplet results.
    """
    if not query or not isinstance(query, str):
        raise ValueError("The query must be a non-empty string.")
    if top_k <= 0:
        raise ValueError("top_k must be a positive integer.")

    if memory_fragment is None:
        memory_fragment = await get_memory_fragment(properties_to_project)

    if collections is None:
        collections = [
            "Entity_name",
            "TextSummary_text",
            "EntityType_name",
            "DocumentChunk_text",
        ]

    try:
        vector_engine = get_vector_engine()
    except Exception as e:
        logger.error("Failed to initialize vector engine: %s", e)
        raise RuntimeError("Initialization error") from e

    send_telemetry("cognee.brute_force_triplet_search EXECUTION STARTED", user.id)

    try:
        results = await asyncio.gather(
            *[
                vector_engine.get_distance_from_collection_elements(collection, query_text=query)
                for collection in collections
            ]
        )

        if all(not item for item in results):
            raise CollectionDistancesNotFoundError()

        node_distances = {collection: result for collection, result in zip(collections, results)}

        await memory_fragment.map_vector_distances_to_graph_nodes(node_distances=node_distances)
        await memory_fragment.map_vector_distances_to_graph_edges(vector_engine, query)

        results = await memory_fragment.calculate_top_triplet_importances(k=top_k)

        send_telemetry("cognee.brute_force_triplet_search EXECUTION COMPLETED", user.id)

        return results

    except Exception as error:
        logger.error(
            "Error during brute force search for user: %s, query: %s. Error: %s",
            user.id,
            query,
            error,
        )
        send_telemetry(
            "cognee.brute_force_triplet_search EXECUTION FAILED", user.id, {"error": str(error)}
        )
        raise error



from typing import Union, BinaryIO, List
from cognee.modules.ingestion import classify
from cognee.infrastructure.databases.relational import get_relational_engine
from sqlalchemy import select
from sqlalchemy.sql import delete as sql_delete
from cognee.modules.data.models import Data, DatasetData, Dataset
from cognee.infrastructure.databases.graph import get_graph_engine
from io import StringIO, BytesIO
import hashlib
import asyncio
from uuid import UUID
from cognee.infrastructure.databases.vector import get_vector_engine
from cognee.infrastructure.engine import DataPoint
from cognee.modules.graph.utils.convert_node_to_data_point import get_all_subclasses
from .exceptions import DocumentNotFoundError, DatasetNotFoundError, DocumentSubgraphNotFoundError
from cognee.shared.logging_utils import get_logger

logger = get_logger()


def get_text_content_hash(text: str) -> str:
    encoded_text = text.encode("utf-8")
    return hashlib.md5(encoded_text).hexdigest()


async def delete(
    data: Union[BinaryIO, List[BinaryIO], str, List[str]],
    dataset_name: str = "main_dataset",
    mode: str = "soft",
):
    """Delete a document and all its related nodes from both relational and graph databases.

    Args:
        data: The data to delete (file, URL, or text)
        dataset_name: Name of the dataset to delete from
        mode: "soft" (default) or "hard" - hard mode also deletes degree-one entity nodes
    """

    # Handle different input types
    if isinstance(data, str):
        if data.startswith("file://"):  # It's a file path
            with open(data.replace("file://", ""), mode="rb") as file:
                classified_data = classify(file)
                content_hash = classified_data.get_metadata()["content_hash"]
                return await delete_single_document(content_hash, dataset_name, mode)
        elif data.startswith("http"):  # It's a URL
            import requests

            response = requests.get(data)
            response.raise_for_status()
            file_data = BytesIO(response.content)
            classified_data = classify(file_data)
            content_hash = classified_data.get_metadata()["content_hash"]
            return await delete_single_document(content_hash, dataset_name, mode)
        else:  # It's a text string
            content_hash = get_text_content_hash(data)
            classified_data = classify(data)
            return await delete_single_document(content_hash, dataset_name, mode)
    elif isinstance(data, list):
        # Handle list of inputs sequentially
        results = []
        for item in data:
            result = await delete(item, dataset_name, mode)
            results.append(result)
        return {"status": "success", "message": "Multiple documents deleted", "results": results}
    else:  # It's already a BinaryIO
        data.seek(0)  # Ensure we're at the start of the file
        classified_data = classify(data)
        content_hash = classified_data.get_metadata()["content_hash"]
        return await delete_single_document(content_hash, dataset_name, mode)


async def delete_single_document(content_hash: str, dataset_name: str, mode: str = "soft"):
    """Delete a single document by its content hash."""

    # Delete from graph database
    deletion_result = await delete_document_subgraph(content_hash, mode)

    logger.info(f"Deletion result: {deletion_result}")

    # Get the deleted node IDs and convert to UUID
    deleted_node_ids = []
    for node_id in deletion_result["deleted_node_ids"]:
        try:
            # Handle both string and UUID formats
            if isinstance(node_id, str):
                # Remove any hyphens if present
                node_id = node_id.replace("-", "")
                deleted_node_ids.append(UUID(node_id))
            else:
                deleted_node_ids.append(node_id)
        except Exception as e:
            logger.error(f"Error converting node ID {node_id} to UUID: {e}")
            continue

    # Delete from vector database
    vector_engine = get_vector_engine()

    # Determine vector collections dynamically
    subclasses = get_all_subclasses(DataPoint)
    vector_collections = []

    for subclass in subclasses:
        index_fields = subclass.model_fields["metadata"].default.get("index_fields", [])
        for field_name in index_fields:
            vector_collections.append(f"{subclass.__name__}_{field_name}")

    # If no collections found, use default collections
    if not vector_collections:
        vector_collections = [
            "DocumentChunk_text",
            "EdgeType_relationship_name",
            "EntityType_name",
            "Entity_name",
            "TextDocument_name",
            "TextSummary_text",
        ]

    # Delete records from each vector collection that exists
    for collection in vector_collections:
        if await vector_engine.has_collection(collection):
            await vector_engine.delete_data_points(
                collection, [str(node_id) for node_id in deleted_node_ids]
            )

    # Delete from relational database
    db_engine = get_relational_engine()
    async with db_engine.get_async_session() as session:
        # Update graph_relationship_ledger with deleted_at timestamps
        from sqlalchemy import update, and_, or_
        from datetime import datetime
        from cognee.modules.data.models.graph_relationship_ledger import GraphRelationshipLedger

        update_stmt = (
            update(GraphRelationshipLedger)
            .where(
                or_(
                    GraphRelationshipLedger.source_node_id.in_(deleted_node_ids),
                    GraphRelationshipLedger.destination_node_id.in_(deleted_node_ids),
                )
            )
            .values(deleted_at=datetime.now())
        )
        await session.execute(update_stmt)

        # Get the data point
        data_point = (
            await session.execute(select(Data).filter(Data.content_hash == content_hash))
        ).scalar_one_or_none()

        if data_point is None:
            raise DocumentNotFoundError(
                f"Document not found in relational DB with content hash: {content_hash}"
            )

        doc_id = data_point.id

        # Get the dataset
        dataset = (
            await session.execute(select(Dataset).filter(Dataset.name == dataset_name))
        ).scalar_one_or_none()

        if dataset is None:
            raise DatasetNotFoundError(f"Dataset not found: {dataset_name}")

        # Delete from dataset_data table
        dataset_delete_stmt = sql_delete(DatasetData).where(
            DatasetData.data_id == doc_id, DatasetData.dataset_id == dataset.id
        )
        await session.execute(dataset_delete_stmt)

        # Check if the document is in any other datasets
        remaining_datasets = (
            await session.execute(select(DatasetData).filter(DatasetData.data_id == doc_id))
        ).scalar_one_or_none()

        # If the document is not in any other datasets, delete it from the data table
        if remaining_datasets is None:
            data_delete_stmt = sql_delete(Data).where(Data.id == doc_id)
            await session.execute(data_delete_stmt)

        await session.commit()

    return {
        "status": "success",
        "message": "Document deleted from both graph and relational databases",
        "graph_deletions": deletion_result["deleted_counts"],
        "content_hash": content_hash,
        "dataset": dataset_name,
        "deleted_node_ids": [
            str(node_id) for node_id in deleted_node_ids
        ],  # Convert back to strings for response
    }


async def delete_document_subgraph(content_hash: str, mode: str = "soft"):
    """Delete a document and all its related nodes in the correct order."""
    graph_db = await get_graph_engine()
    subgraph = await graph_db.get_document_subgraph(content_hash)
    if not subgraph:
        raise DocumentSubgraphNotFoundError(f"Document not found with content hash: {content_hash}")

    # Delete in the correct order to maintain graph integrity
    deletion_order = [
        ("orphan_entities", "orphaned entities"),
        ("orphan_types", "orphaned entity types"),
        (
            "made_from_nodes",
            "made_from nodes",
        ),  # Move before chunks since summaries are connected to chunks
        ("chunks", "document chunks"),
        ("document", "document"),
    ]

    deleted_counts = {}
    deleted_node_ids = []
    for key, description in deletion_order:
        nodes = subgraph[key]
        if nodes:
            for node in nodes:
                node_id = node["id"]
                await graph_db.delete_node(node_id)
                deleted_node_ids.append(node_id)
            deleted_counts[description] = len(nodes)

    # If hard mode, also delete degree-one nodes
    if mode == "hard":
        # Get and delete degree one entity nodes
        degree_one_entity_nodes = await graph_db.get_degree_one_nodes("Entity")
        for node in degree_one_entity_nodes:
            await graph_db.delete_node(node["id"])
            deleted_node_ids.append(node["id"])
            deleted_counts["degree_one_entities"] = deleted_counts.get("degree_one_entities", 0) + 1

        # Get and delete degree one entity types
        degree_one_entity_types = await graph_db.get_degree_one_nodes("EntityType")
        for node in degree_one_entity_types:
            await graph_db.delete_node(node["id"])
            deleted_node_ids.append(node["id"])
            deleted_counts["degree_one_types"] = deleted_counts.get("degree_one_types", 0) + 1

    return {
        "status": "success",
        "deleted_counts": deleted_counts,
        "content_hash": content_hash,
        "deleted_node_ids": deleted_node_ids,
    }



from typing import Any, Optional
from collections import Counter
import string

from cognee.infrastructure.engine import DataPoint
from cognee.modules.graph.exceptions.exceptions import EntityNotFoundError
from cognee.modules.graph.utils.convert_node_to_data_point import get_all_subclasses
from cognee.modules.retrieval.base_retriever import BaseRetriever
from cognee.modules.retrieval.utils.brute_force_triplet_search import brute_force_triplet_search
from cognee.modules.retrieval.utils.completion import generate_completion
from cognee.modules.retrieval.utils.stop_words import DEFAULT_STOP_WORDS


class GraphCompletionRetriever(BaseRetriever):
    """Retriever for handling graph-based completion searches."""

    def __init__(
        self,
        user_prompt_path: str = "graph_context_for_question.txt",
        system_prompt_path: str = "answer_simple_question.txt",
        top_k: Optional[int] = 5,
    ):
        """Initialize retriever with prompt paths and search parameters."""
        self.user_prompt_path = user_prompt_path
        self.system_prompt_path = system_prompt_path
        self.top_k = top_k if top_k is not None else 5

    def _get_nodes(self, retrieved_edges: list) -> dict:
        """Creates a dictionary of nodes with their names and content."""
        nodes = {}
        for edge in retrieved_edges:
            for node in (edge.node1, edge.node2):
                if node.id not in nodes:
                    text = node.attributes.get("text")
                    if text:
                        name = self._get_title(text)
                        content = text
                    else:
                        name = node.attributes.get("name", "Unnamed Node")
                        content = name
                    nodes[node.id] = {"node": node, "name": name, "content": content}
        return nodes

    async def resolve_edges_to_text(self, retrieved_edges: list) -> str:
        """Converts retrieved graph edges into a human-readable string format."""
        nodes = self._get_nodes(retrieved_edges)
        node_section = "\n".join(
            f"Node: {info['name']}\n__node_content_start__\n{info['content']}\n__node_content_end__\n"
            for info in nodes.values()
        )
        connection_section = "\n".join(
            f"{nodes[edge.node1.id]['name']} --[{edge.attributes['relationship_type']}]--> {nodes[edge.node2.id]['name']}"
            for edge in retrieved_edges
        )
        return f"Nodes:\n{node_section}\n\nConnections:\n{connection_section}"

    async def get_triplets(self, query: str) -> list:
        """Retrieves relevant graph triplets."""
        subclasses = get_all_subclasses(DataPoint)
        vector_index_collections = []

        for subclass in subclasses:
            if "metadata" in subclass.model_fields:
                metadata_field = subclass.model_fields["metadata"]
                if hasattr(metadata_field, "default") and metadata_field.default is not None:
                    if isinstance(metadata_field.default, dict):
                        index_fields = metadata_field.default.get("index_fields", [])
                        for field_name in index_fields:
                            vector_index_collections.append(f"{subclass.__name__}_{field_name}")

        found_triplets = await brute_force_triplet_search(
            query, top_k=self.top_k, collections=vector_index_collections or None
        )

        return found_triplets

    async def get_context(self, query: str) -> str:
        """Retrieves and resolves graph triplets into context."""
        try:
            triplets = await self.get_triplets(query)
        except EntityNotFoundError:
            return ""

        if len(triplets) == 0:
            return ""

        return await self.resolve_edges_to_text(triplets)

    async def get_completion(self, query: str, context: Optional[Any] = None) -> Any:
        """Generates a completion using graph connections context."""
        if context is None:
            context = await self.get_context(query)

        completion = await generate_completion(
            query=query,
            context=context,
            user_prompt_path=self.user_prompt_path,
            system_prompt_path=self.system_prompt_path,
        )
        return [completion]

    def _top_n_words(self, text, stop_words=None, top_n=3, separator=", "):
        """Concatenates the top N frequent words in text."""
        if stop_words is None:
            stop_words = DEFAULT_STOP_WORDS

        words = [word.lower().strip(string.punctuation) for word in text.split()]

        if stop_words:
            words = [word for word in words if word and word not in stop_words]

        top_words = [word for word, freq in Counter(words).most_common(top_n)]

        return separator.join(top_words)

    def _get_title(self, text: str, first_n_words: int = 7, top_n_words: int = 3) -> str:
        """Creates a title, by combining first words with most frequent words from the text."""
        first_n_words = text.split()[:first_n_words]
        top_n_words = self._top_n_words(text, top_n=top_n_words)
        return f"{' '.join(first_n_words)}... [{top_n_words}]"



from cognee.shared.logging_utils import get_logger
from typing import List

from pydantic import BaseModel

from cognee.infrastructure.entities.BaseEntityExtractor import BaseEntityExtractor
from cognee.modules.engine.models import Entity
from cognee.modules.engine.models.EntityType import EntityType
from cognee.infrastructure.llm.prompts import read_query_prompt, render_prompt
from cognee.infrastructure.llm.get_llm_client import get_llm_client

logger = get_logger("llm_entity_extractor")


class EntityList(BaseModel):
    """Response model containing a list of extracted entities."""

    entities: List[Entity]


class LLMEntityExtractor(BaseEntityExtractor):
    """Entity extractor that uses an LLM to identify entities in text."""

    def __init__(
        self,
        system_prompt_template: str = "extract_entities_system.txt",
        user_prompt_template: str = "extract_entities_user.txt",
    ):
        """Initialize the LLM entity extractor."""
        self.system_prompt_template = system_prompt_template
        self.user_prompt_template = user_prompt_template
        self._entity_type_cache = {}

    def _get_entity_type(self, type_name: str) -> EntityType:
        """Get or create an EntityType object."""
        type_name = type_name.upper()

        if type_name not in self._entity_type_cache:
            self._entity_type_cache[type_name] = EntityType(
                name=type_name, description=f"Entity type for {type_name.lower()} entities"
            )

        return self._entity_type_cache[type_name]

    async def extract_entities(self, text: str) -> List[Entity]:
        """Extract entities from text using an LLM."""
        if not text or not isinstance(text, str):
            logger.warning("Invalid input text for entity extraction")
            return []

        try:
            logger.info(f"Extracting entities from text: {text[:100]}...")

            llm_client = get_llm_client()
            user_prompt = render_prompt(self.user_prompt_template, {"text": text})
            system_prompt = read_query_prompt(self.system_prompt_template)

            response = await llm_client.acreate_structured_output(
                text_input=user_prompt,
                system_prompt=system_prompt,
                response_model=EntityList,
            )

            if not response.entities:
                logger.warning("No entities were extracted from the text")
                return []

            logger.info(f"Extracted {len(response.entities)} entities")
            return response.entities

        except Exception as e:
            logger.error(f"Entity extraction failed: {str(e)}")
            return []



from enum import Enum
from pydantic import BaseModel
from cognee.infrastructure.databases.vector import get_vectordb_config
from cognee.infrastructure.llm import get_llm_config


class ConfigChoice(BaseModel):
    value: str
    label: str


class ModelName(Enum):
    openai = "openai"
    ollama = "ollama"
    anthropic = "anthropic"
    gemini = "gemini"


class LLMConfig(BaseModel):
    api_key: str
    model: str
    provider: str
    endpoint: str
    api_version: str
    models: dict[str, list[ConfigChoice]]
    providers: list[ConfigChoice]


class VectorDBConfig(BaseModel):
    api_key: str
    url: str
    provider: str
    providers: list[ConfigChoice]


class SettingsDict(BaseModel):
    llm: LLMConfig
    vector_db: VectorDBConfig


def get_settings() -> SettingsDict:
    llm_config = get_llm_config()

    vector_dbs = [
        {
            "value": "weaviate",
            "label": "Weaviate",
        },
        {
            "value": "qdrant",
            "label": "Qdrant",
        },
        {
            "value": "lancedb",
            "label": "LanceDB",
        },
        {
            "value": "pgvector",
            "label": "PGVector",
        },
    ]

    vector_config = get_vectordb_config()

    llm_providers = [
        {
            "value": "openai",
            "label": "OpenAI",
        },
        {
            "value": "ollama",
            "label": "Ollama",
        },
        {
            "value": "anthropic",
            "label": "Anthropic",
        },
        {
            "value": "gemini",
            "label": "Gemini",
        },
    ]

    return SettingsDict.model_validate(
        dict(
            llm={
                "provider": llm_config.llm_provider,
                "model": llm_config.llm_model,
                "endpoint": llm_config.llm_endpoint,
                "api_version": llm_config.llm_api_version,
                "api_key": (llm_config.llm_api_key[0:10] + "*" * (len(llm_config.llm_api_key) - 10))
                if llm_config.llm_api_key
                else None,
                "providers": llm_providers,
                "models": {
                    "openai": [
                        {
                            "value": "gpt-4o-mini",
                            "label": "gpt-4o-mini",
                        },
                        {
                            "value": "gpt-4o",
                            "label": "gpt-4o",
                        },
                        {
                            "value": "gpt-4-turbo",
                            "label": "gpt-4-turbo",
                        },
                        {
                            "value": "gpt-3.5-turbo",
                            "label": "gpt-3.5-turbo",
                        },
                    ],
                    "ollama": [
                        {
                            "value": "llama3",
                            "label": "llama3",
                        },
                        {
                            "value": "mistral",
                            "label": "mistral",
                        },
                    ],
                    "anthropic": [
                        {
                            "value": "Claude 3 Opus",
                            "label": "Claude 3 Opus",
                        },
                        {
                            "value": "Claude 3 Sonnet",
                            "label": "Claude 3 Sonnet",
                        },
                        {
                            "value": "Claude 3 Haiku",
                            "label": "Claude 3 Haiku",
                        },
                    ],
                    "gemini": [
                        {
                            "value": "gemini-2.0-flash-exp",
                            "label": "Gemini 2.0 Flash",
                        },
                    ],
                },
            },
            vector_db={
                "provider": vector_config.vector_db_provider,
                "url": vector_config.vector_db_url,
                "api_key": (
                    vector_config.vector_db_key[0:10]
                    + "*" * (len(vector_config.vector_db_key) - 10)
                ),
                "providers": vector_dbs,
            },
        )
    )



import time
import asyncio
from cognee.shared.logging_utils import get_logger
from cognee.infrastructure.llm.rate_limiter import (
    sleep_and_retry_sync,
    sleep_and_retry_async,
    is_rate_limit_error,
)

logger = get_logger()


# Test function to be decorated
@sleep_and_retry_sync(max_retries=3, initial_backoff=0.1, backoff_factor=2.0)
def test_function_sync():
    """A test function that raises rate limit errors a few times, then succeeds."""
    if hasattr(test_function_sync, "counter"):
        test_function_sync.counter += 1
    else:
        test_function_sync.counter = 1

    if test_function_sync.counter <= 2:
        error_msg = "429 Too Many Requests: Rate limit exceeded"
        logger.info(f"Attempt {test_function_sync.counter}: Raising rate limit error")
        raise Exception(error_msg)

    logger.info(f"Attempt {test_function_sync.counter}: Success!")
    return f"Success on attempt {test_function_sync.counter}"


# Test async function to be decorated
@sleep_and_retry_async(max_retries=3, initial_backoff=0.1, backoff_factor=2.0)
async def test_function_async():
    """An async test function that raises rate limit errors a few times, then succeeds."""
    if hasattr(test_function_async, "counter"):
        test_function_async.counter += 1
    else:
        test_function_async.counter = 1

    if test_function_async.counter <= 2:
        error_msg = "429 Too Many Requests: Rate limit exceeded"
        logger.info(f"Attempt {test_function_async.counter}: Raising rate limit error")
        raise Exception(error_msg)

    logger.info(f"Attempt {test_function_async.counter}: Success!")
    return f"Success on attempt {test_function_async.counter}"


def test_is_rate_limit_error():
    """Test the rate limit error detection function."""
    print("\n=== Testing Rate Limit Error Detection ===")

    # Test various error messages that should be detected as rate limit errors
    rate_limit_errors = [
        "429 Rate limit exceeded",
        "Too many requests",
        "rate_limit_exceeded",
        "ratelimit error",
        "You have exceeded your quota",
        "capacity has been exceeded",
        "Service throttled",
    ]

    # Test error messages that should not be detected as rate limit errors
    non_rate_limit_errors = [
        "404 Not Found",
        "500 Internal Server Error",
        "Invalid API Key",
        "Bad Request",
    ]

    # Check that rate limit errors are correctly identified
    for error in rate_limit_errors:
        error_obj = Exception(error)
        result = is_rate_limit_error(error_obj)
        print(f"Error '{error}': {'✓' if result else '✗'} {result}")
        assert result, f"Failed to identify rate limit error: {error}"
        print(f"✓ Correctly identified as rate limit error: {error}")

    # Check that non-rate limit errors are not misidentified
    for error in non_rate_limit_errors:
        error_obj = Exception(error)
        result = is_rate_limit_error(error_obj)
        print(f"Error '{error}': {'✓' if not result else '✗'} {not result}")
        assert not result, f"Incorrectly identified as rate limit error: {error}"
        print(f"✓ Correctly identified as non-rate limit error: {error}")

    print("✅ PASS: Rate limit error detection is working correctly")


def test_sync_retry():
    """Test the synchronous retry decorator."""
    print("\n=== Testing Synchronous Sleep and Retry ===")

    # Reset counter for the test function
    if hasattr(test_function_sync, "counter"):
        del test_function_sync.counter

    # Time the execution to verify backoff is working
    start_time = time.time()

    try:
        result = test_function_sync()
        end_time = time.time()
        elapsed = end_time - start_time

        # Verify results
        print(f"Result: {result}")
        print(f"Test completed in {elapsed:.2f} seconds")
        print(f"Number of attempts: {test_function_sync.counter}")

        # The function should succeed on the 3rd attempt (after 2 failures)
        assert test_function_sync.counter == 3, (
            f"Expected 3 attempts, got {test_function_sync.counter}"
        )
        assert elapsed >= 0.3, f"Expected at least 0.3 seconds of backoff, got {elapsed:.2f}"

        print("✅ PASS: Synchronous retry mechanism is working correctly")
    except Exception as e:
        print(f"❌ FAIL: Test encountered an unexpected error: {str(e)}")
        raise


async def test_async_retry():
    """Test the asynchronous retry decorator."""
    print("\n=== Testing Asynchronous Sleep and Retry ===")

    # Reset counter for the test function
    if hasattr(test_function_async, "counter"):
        del test_function_async.counter

    # Time the execution to verify backoff is working
    start_time = time.time()

    try:
        result = await test_function_async()
        end_time = time.time()
        elapsed = end_time - start_time

        # Verify results
        print(f"Result: {result}")
        print(f"Test completed in {elapsed:.2f} seconds")
        print(f"Number of attempts: {test_function_async.counter}")

        # The function should succeed on the 3rd attempt (after 2 failures)
        assert test_function_async.counter == 3, (
            f"Expected 3 attempts, got {test_function_async.counter}"
        )
        assert elapsed >= 0.3, f"Expected at least 0.3 seconds of backoff, got {elapsed:.2f}"

        print("✅ PASS: Asynchronous retry mechanism is working correctly")
    except Exception as e:
        print(f"❌ FAIL: Test encountered an unexpected error: {str(e)}")
        raise


async def test_retry_max_exceeded():
    """Test what happens when max retries is exceeded."""
    print("\n=== Testing Max Retries Exceeded ===")

    @sleep_and_retry_async(max_retries=2, initial_backoff=0.1)
    async def always_fails():
        """A function that always raises a rate limit error."""
        error_msg = "429 Too Many Requests: Rate limit always exceeded"
        logger.info(f"Always fails with: {error_msg}")
        raise Exception(error_msg)

    try:
        # This should fail after 2 retries (3 attempts total)
        await always_fails()
        print("❌ FAIL: Function should have failed but succeeded")
    except Exception as e:
        print(f"Expected error after max retries: {str(e)}")
        print("✅ PASS: Function correctly failed after max retries exceeded")


async def main():
    """Run all the retry tests."""
    test_is_rate_limit_error()
    test_sync_retry()
    await test_async_retry()
    await test_retry_max_exceeded()

    print("\n=== All Rate Limiting Retry Tests Complete ===")


if __name__ == "__main__":
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        loop.run_until_complete(main())
    finally:
        loop.run_until_complete(loop.shutdown_asyncgens())



from fastapi import Form, UploadFile, Depends
from fastapi.responses import JSONResponse
from fastapi import APIRouter
from typing import List, Optional
import subprocess
from cognee.shared.logging_utils import get_logger
import requests
from cognee.modules.users.models import User
from cognee.modules.users.methods import get_authenticated_user

logger = get_logger()


def get_delete_router() -> APIRouter:
    router = APIRouter()

    @router.delete("/", response_model=None)
    async def delete(
        data: List[UploadFile],
        dataset_name: str = Form("main_dataset"),
        mode: str = Form("soft"),
        user: User = Depends(get_authenticated_user),
    ):
        """This endpoint is responsible for deleting data from the graph.

        Args:
            data: The data to delete (files, URLs, or text)
            dataset_name: Name of the dataset to delete from (default: "main_dataset")
            mode: "soft" (default) or "hard" - hard mode also deletes degree-one entity nodes
            user: Authenticated user
        """
        from cognee.api.v1.delete import delete as cognee_delete

        try:
            # Handle each file in the list
            results = []
            for file in data:
                if file.filename.startswith("http"):
                    if "github" in file.filename:
                        # For GitHub repos, we need to get the content hash of each file
                        repo_name = file.filename.split("/")[-1].replace(".git", "")
                        subprocess.run(
                            ["git", "clone", file.filename, f".data/{repo_name}"], check=True
                        )
                        # Note: This would need to be implemented to get content hashes of all files
                        # For now, we'll just return an error
                        return JSONResponse(
                            status_code=400,
                            content={"error": "Deleting GitHub repositories is not yet supported"},
                        )
                    else:
                        # Fetch and delete the data from other types of URL
                        response = requests.get(file.filename)
                        response.raise_for_status()
                        file_data = response.content
                        result = await cognee_delete(
                            file_data, dataset_name=dataset_name, mode=mode
                        )
                        results.append(result)
                else:
                    # Handle uploaded file
                    result = await cognee_delete(file, dataset_name=dataset_name, mode=mode)
                    results.append(result)

            if len(results) == 1:
                return results[0]
            else:
                return {
                    "status": "success",
                    "message": "Multiple documents deleted",
                    "results": results,
                }
        except Exception as error:
            logger.error(f"Error during deletion: {str(error)}")
            return JSONResponse(status_code=409, content={"error": str(error)})

    return router



from cognee.shared.logging_utils import get_logger
from typing import List, Dict, Union

from cognee.exceptions import InvalidValueError
from cognee.modules.graph.exceptions import EntityNotFoundError, EntityAlreadyExistsError
from cognee.infrastructure.databases.graph.graph_db_interface import GraphDBInterface
from cognee.modules.graph.cognee_graph.CogneeGraphElements import Node, Edge
from cognee.modules.graph.cognee_graph.CogneeAbstractGraph import CogneeAbstractGraph
import heapq

logger = get_logger()


class CogneeGraph(CogneeAbstractGraph):
    """
    Concrete implementation of the AbstractGraph class for Cognee.

    This class provides the functionality to manage nodes and edges,
    and project a graph from a database using adapters.
    """

    nodes: Dict[str, Node]
    edges: List[Edge]
    directed: bool

    def __init__(self, directed: bool = True):
        self.nodes = {}
        self.edges = []
        self.directed = directed

    def add_node(self, node: Node) -> None:
        if node.id not in self.nodes:
            self.nodes[node.id] = node
        else:
            raise EntityAlreadyExistsError(message=f"Node with id {node.id} already exists.")

    def add_edge(self, edge: Edge) -> None:
        self.edges.append(edge)
        edge.node1.add_skeleton_edge(edge)
        edge.node2.add_skeleton_edge(edge)

    def get_node(self, node_id: str) -> Node:
        return self.nodes.get(node_id, None)

    def get_edges_from_node(self, node_id: str) -> List[Edge]:
        node = self.get_node(node_id)
        if node:
            return node.skeleton_edges
        else:
            raise EntityNotFoundError(message=f"Node with id {node_id} does not exist.")

    def get_edges(self) -> List[Edge]:
        return self.edges

    async def project_graph_from_db(
        self,
        adapter: Union[GraphDBInterface],
        node_properties_to_project: List[str],
        edge_properties_to_project: List[str],
        directed=True,
        node_dimension=1,
        edge_dimension=1,
        memory_fragment_filter=[],
    ) -> None:
        if node_dimension < 1 or edge_dimension < 1:
            raise InvalidValueError(message="Dimensions must be positive integers")

        try:
            if len(memory_fragment_filter) == 0:
                nodes_data, edges_data = await adapter.get_graph_data()
            else:
                nodes_data, edges_data = await adapter.get_filtered_graph_data(
                    attribute_filters=memory_fragment_filter
                )

            if not nodes_data:
                raise EntityNotFoundError(message="No node data retrieved from the database.")
            if not edges_data:
                raise EntityNotFoundError(message="No edge data retrieved from the database.")

            for node_id, properties in nodes_data:
                node_attributes = {key: properties.get(key) for key in node_properties_to_project}
                self.add_node(Node(str(node_id), node_attributes, dimension=node_dimension))

            for source_id, target_id, relationship_type, properties in edges_data:
                source_node = self.get_node(str(source_id))
                target_node = self.get_node(str(target_id))
                if source_node and target_node:
                    edge_attributes = {
                        key: properties.get(key) for key in edge_properties_to_project
                    }
                    edge_attributes["relationship_type"] = relationship_type

                    edge = Edge(
                        source_node,
                        target_node,
                        attributes=edge_attributes,
                        directed=directed,
                        dimension=edge_dimension,
                    )
                    self.add_edge(edge)

                    source_node.add_skeleton_edge(edge)
                    target_node.add_skeleton_edge(edge)

                else:
                    raise EntityNotFoundError(
                        message=f"Edge references nonexistent nodes: {source_id} -> {target_id}"
                    )

        except (ValueError, TypeError) as e:
            print(f"Error projecting graph: {e}")
            raise e

    async def map_vector_distances_to_graph_nodes(self, node_distances) -> None:
        for category, scored_results in node_distances.items():
            for scored_result in scored_results:
                node_id = str(scored_result.id)
                score = scored_result.score
                node = self.get_node(node_id)
                if node:
                    node.add_attribute("vector_distance", score)

    async def map_vector_distances_to_graph_edges(self, vector_engine, query) -> None:
        try:
            query_vector = await vector_engine.embed_data([query])
            query_vector = query_vector[0]
            if query_vector is None or len(query_vector) == 0:
                raise ValueError("Failed to generate query embedding.")

            edge_distances = await vector_engine.get_distance_from_collection_elements(
                "EdgeType_relationship_name", query_text=query
            )

            embedding_map = {result.payload["text"]: result.score for result in edge_distances}

            for edge in self.edges:
                relationship_type = edge.attributes.get("relationship_type")
                if not relationship_type or relationship_type not in embedding_map:
                    print(f"Edge {edge} has an unknown or missing relationship type.")
                    continue

                edge.attributes["vector_distance"] = embedding_map[relationship_type]

        except Exception as ex:
            print(f"Error mapping vector distances to edges: {ex}")
            raise ex

    async def calculate_top_triplet_importances(self, k: int) -> List:
        min_heap = []
        for i, edge in enumerate(self.edges):
            source_node = self.get_node(edge.node1.id)
            target_node = self.get_node(edge.node2.id)

            source_distance = source_node.attributes.get("vector_distance", 1) if source_node else 1
            target_distance = target_node.attributes.get("vector_distance", 1) if target_node else 1
            edge_distance = edge.attributes.get("vector_distance", 1)

            total_distance = source_distance + target_distance + edge_distance

            heapq.heappush(min_heap, (-total_distance, i, edge))
            if len(min_heap) > k:
                heapq.heappop(min_heap)

        return [edge for _, _, edge in sorted(min_heap)]



from typing import List, Any, Union

from ..tokenizer_interface import TokenizerInterface


class GeminiTokenizer(TokenizerInterface):
    def __init__(
        self,
        model: str,
        max_tokens: int = 3072,
    ):
        self.model = model
        self.max_tokens = max_tokens

        # Get LLM API key from config
        from cognee.infrastructure.databases.vector.embeddings.config import get_embedding_config
        from cognee.infrastructure.llm.config import get_llm_config

        config = get_embedding_config()
        llm_config = get_llm_config()

        import google.generativeai as genai

        genai.configure(api_key=config.embedding_api_key or llm_config.llm_api_key)

    def extract_tokens(self, text: str) -> List[Any]:
        raise NotImplementedError

    def decode_single_token(self, encoding: int):
        # Gemini tokenizer doesn't have the option to decode tokens
        raise NotImplementedError

    def count_tokens(self, text: str) -> int:
        """
        Returns the number of tokens in the given text.
        Args:
            text: str

        Returns:
            number of tokens in the given text

        """
        import google.generativeai as genai

        return len(genai.embed_content(model=f"models/{self.model}", content=text))



from llama_index.core import Document
from llama_index.core.schema import ImageDocument
from cognee.modules.ingestion import save_data_to_file
from typing import Union


def get_data_from_llama_index(data_point: Union[Document, ImageDocument], dataset_name: str) -> str:
    # Specific type checking is used to ensure it's not a child class from Document
    if isinstance(data_point, Document) and type(data_point) is Document:
        file_path = data_point.metadata.get("file_path")
        if file_path is None:
            file_path = save_data_to_file(data_point.text)
            return file_path
        return file_path
    elif isinstance(data_point, ImageDocument) and type(data_point) is ImageDocument:
        if data_point.image_path is None:
            file_path = save_data_to_file(data_point.text)
            return file_path
        return data_point.image_path



from cognee.shared.logging_utils import get_logger
import json
from typing import List
from cognee.eval_framework.evaluation.evaluation_executor import EvaluationExecutor
from cognee.eval_framework.analysis.metrics_calculator import calculate_metrics_statistics
from cognee.eval_framework.analysis.dashboard_generator import create_dashboard
from cognee.infrastructure.files.storage import LocalStorage
from cognee.infrastructure.databases.relational.get_relational_engine import (
    get_relational_engine,
    get_relational_config,
)
from cognee.modules.data.models.metrics_data import Metrics
from cognee.modules.data.models.metrics_base import MetricsBase


logger = get_logger()


async def create_and_insert_metrics_table(questions_payload):
    relational_config = get_relational_config()
    relational_engine = get_relational_engine()

    if relational_engine.engine.dialect.name == "sqlite":
        LocalStorage.ensure_directory_exists(relational_config.db_path)

    async with relational_engine.engine.begin() as connection:
        if len(MetricsBase.metadata.tables.keys()) > 0:
            await connection.run_sync(MetricsBase.metadata.create_all)

    async with relational_engine.get_async_session() as session:
        data_point = Metrics(payload=questions_payload)
        session.add(data_point)
        await session.commit()


async def execute_evaluation(params: dict) -> None:
    """Execute the evaluation step and save results."""
    logger.info("Evaluation started...")
    try:
        with open(params["answers_path"], "r", encoding="utf-8") as f:
            answers = json.load(f)
    except FileNotFoundError:
        raise FileNotFoundError(f"Could not find the file: {params['answers_path']}")
    except json.JSONDecodeError as e:
        raise ValueError(f"Error decoding JSON from {params['answers_path']}: {e}")

    logger.info(f"Loaded {len(answers)} answers from {params['answers_path']}")
    evaluator = EvaluationExecutor(
        evaluator_engine=params["evaluation_engine"],
        evaluate_contexts=params["evaluating_contexts"],
    )
    metrics = await evaluator.execute(
        answers=answers, evaluator_metrics=params["evaluation_metrics"]
    )
    with open(params["metrics_path"], "w", encoding="utf-8") as f:
        json.dump(metrics, f, ensure_ascii=False, indent=4)

    await create_and_insert_metrics_table(metrics)
    logger.info("Evaluation completed")
    return metrics


async def run_evaluation(params: dict) -> List[dict]:
    """Run each step of the evaluation pipeline based on configuration flags."""
    # Step 1: Evaluate answers if requested
    if params.get("evaluating_answers"):
        metrics = await execute_evaluation(params)
    else:
        logger.info("Skipping evaluation as evaluating_answers is False")

    # Step 2: Calculate metrics if requested
    if params.get("calculate_metrics"):
        logger.info("Calculating metrics statistics...")
        calculate_metrics_statistics(
            json_data=params["metrics_path"], aggregate_output_path=params["aggregate_metrics_path"]
        )
        logger.info("Metrics calculation completed")
        return metrics
    else:
        logger.info("Skipping metrics calculation as calculate_metrics is False")
        return []



from datetime import datetime
from uuid import UUID


def serialize_data(data):
    """Recursively convert datetime objects in dictionaries/lists to ISO format."""
    if isinstance(data, dict):
        return {key: serialize_data(value) for key, value in data.items()}
    elif isinstance(data, list):
        return [serialize_data(item) for item in data]
    elif isinstance(data, datetime):
        return data.isoformat()  # Convert datetime to ISO 8601 string
    elif isinstance(data, UUID):
        return str(data)
    else:
        return data



from uuid import UUID as uuid_UUID
from typing import Optional
from sqlalchemy import ForeignKey, Column, UUID
from sqlalchemy.orm import relationship, Mapped
from fastapi_users.db import SQLAlchemyBaseUserTableUUID
from .Principal import Principal
from .UserRole import UserRole
from .Role import Role
from fastapi_users import schemas


class User(SQLAlchemyBaseUserTableUUID, Principal):
    __tablename__ = "users"

    id = Column(UUID, ForeignKey("principals.id", ondelete="CASCADE"), primary_key=True)

    # Foreign key to Tenant (Many-to-One relationship)
    tenant_id = Column(UUID, ForeignKey("tenants.id"))

    # Many-to-Many Relationship with Roles
    roles: Mapped[list["Role"]] = relationship(
        "Role",
        secondary=UserRole.__tablename__,
        back_populates="users",
    )

    # Relationship to Tenant
    tenant = relationship(
        "Tenant",
        back_populates="users",
        foreign_keys=[tenant_id],
    )

    # ACL Relationship (One-to-Many)
    acls = relationship("ACL", back_populates="principal", cascade="all, delete")

    __mapper_args__ = {
        "polymorphic_identity": "user",
    }


# Keep these schemas in sync with User model
class UserRead(schemas.BaseUser[uuid_UUID]):
    tenant_id: Optional[uuid_UUID] = None


class UserCreate(schemas.BaseUserCreate):
    tenant_id: Optional[uuid_UUID] = None


class UserUpdate(schemas.BaseUserUpdate):
    pass



import unittest
import os
import uuid
from unittest.mock import patch, MagicMock
import sys

# Import the telemetry function to test
from cognee.shared.utils import send_telemetry


class TestTelemetry(unittest.TestCase):
    @patch("cognee.shared.utils.requests.post")
    def test_telemetry_enabled(self, mock_post):
        # Setup mock response
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_post.return_value = mock_response

        # Check if .anon_id exists in the project root
        anon_id_path = os.path.join(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__))), ".anon_id"
        )

        if not os.path.exists(anon_id_path):
            # Create the file with a test ID if it doesn't exist
            with open(anon_id_path, "w") as f:
                f.write("test-machine")
            print(f"Created .anon_id file at {anon_id_path}", file=sys.stderr)

        self.assertTrue(os.path.exists(anon_id_path), "The .anon_id file should exist")

        # Verify the file has content
        with open(anon_id_path, "r") as f:
            content = f.read().strip()

        self.assertTrue(len(content) > 0, "The .anon_id file should not be empty")

        # Ensure telemetry is enabled for this test
        if "TELEMETRY_DISABLED" in os.environ:
            del os.environ["TELEMETRY_DISABLED"]

        # Make sure ENV is not test or dev
        original_env = os.environ.get("ENV")
        os.environ["ENV"] = "prod"  # Set to dev to ensure telemetry is sent

        # Generate a random user ID for testing
        test_user_id = str(uuid.uuid4())

        # Test sending telemetry
        event_name = "test_event"
        additional_props = {"test_key": "test_value"}

        send_telemetry(event_name, test_user_id, additional_props)

        # Verify telemetry was sent
        mock_post.assert_called_once()

        # Get the args that were passed to post
        args, kwargs = mock_post.call_args

        # Check that the payload contains our data
        self.assertIn("json", kwargs)
        payload = kwargs["json"]

        # Verify payload contains expected data
        self.assertEqual(payload.get("event_name"), event_name)

        # Check that user_id is in the correct nested structure
        self.assertIn("user_properties", payload)
        self.assertEqual(payload["user_properties"].get("user_id"), str(test_user_id))

        # Also check that user_id is in the properties
        self.assertIn("properties", payload)
        self.assertEqual(payload["properties"].get("user_id"), str(test_user_id))

        # Check that additional properties are included
        self.assertEqual(payload["properties"].get("test_key"), "test_value")

        # Restore original ENV if it existed
        if original_env is not None:
            os.environ["ENV"] = original_env
        else:
            del os.environ["ENV"]

    @patch("cognee.shared.utils.requests.post")
    def test_telemetry_disabled(self, mock_post):
        # Enable the TELEMETRY_DISABLED environment variable
        os.environ["TELEMETRY_DISABLED"] = "1"

        # Test sending telemetry
        send_telemetry("disabled_test", "user123", {"key": "value"})

        # Verify telemetry was not sent
        mock_post.assert_not_called()

        # Clean up
        del os.environ["TELEMETRY_DISABLED"]

    @patch("cognee.shared.utils.requests.post")
    def test_telemetry_dev_env(self, mock_post):
        # Set ENV to dev which should disable telemetry
        original_env = os.environ.get("ENV")
        os.environ["ENV"] = "dev"

        if "TELEMETRY_DISABLED" in os.environ:
            del os.environ["TELEMETRY_DISABLED"]

        # Test sending telemetry
        send_telemetry("dev_test", "user123", {"key": "value"})

        # Verify telemetry was not sent in dev environment
        mock_post.assert_not_called()

        # Restore original ENV if it existed
        if original_env is not None:
            os.environ["ENV"] = original_env
        else:
            del os.environ["ENV"]


if __name__ == "__main__":
    unittest.main()



import os
from functools import lru_cache
from fastapi_users import models
from fastapi_users.jwt import generate_jwt
from fastapi_users.authentication import (
    AuthenticationBackend,
    BearerTransport,
    JWTStrategy,
)

from typing import Optional

from cognee.modules.users.models import User
from cognee.modules.users.methods import get_user


class CustomJWTStrategy(JWTStrategy):
    async def write_token(self, user: User, lifetime_seconds: Optional[int] = None) -> str:
        # JoinLoad tenant and role information to user object
        user = await get_user(user.id)

        if user.tenant:
            data = {"user_id": str(user.id), "tenant_id": str(user.tenant.id), "roles": user.roles}
        else:
            # The default tenant is None
            data = {"user_id": str(user.id), "tenant_id": None, "roles": user.roles}
        return generate_jwt(data, self.encode_key, self.lifetime_seconds, algorithm=self.algorithm)


@lru_cache
def get_auth_backend():
    bearer_transport = BearerTransport(tokenUrl="auth/jwt/login")

    def get_jwt_strategy() -> JWTStrategy[models.UP, models.ID]:
        secret = os.getenv("FASTAPI_USERS_JWT_SECRET", "super_secret")
        return CustomJWTStrategy(secret, lifetime_seconds=3600)

    auth_backend = AuthenticationBackend(
        name="jwt",
        transport=bearer_transport,
        get_strategy=get_jwt_strategy,
    )

    return auth_backend



from typing import List, Any
import tiktoken

from ..tokenizer_interface import TokenizerInterface


class TikTokenTokenizer(TokenizerInterface):
    """
    Tokenizer adapter for OpenAI.
    Inteded to be used as part of LLM Embedding and LLM Adapters classes
    """

    def __init__(
        self,
        model: str,
        max_tokens: int = 8191,
    ):
        self.model = model
        self.max_tokens = max_tokens
        # Initialize TikToken for GPT based on model
        self.tokenizer = tiktoken.encoding_for_model(self.model)

    def extract_tokens(self, text: str) -> List[Any]:
        # Using TikToken's method to tokenize text
        token_ids = self.tokenizer.encode(text)
        return token_ids

    def decode_token_list(self, tokens: List[Any]) -> List[Any]:
        if not isinstance(tokens, list):
            tokens = [tokens]
        return [self.tokenizer.decode(i) for i in tokens]

    def decode_single_token(self, token: int):
        return self.tokenizer.decode_single_token_bytes(token).decode("utf-8", errors="replace")

    def count_tokens(self, text: str) -> int:
        """
        Returns the number of tokens in the given text.
        Args:
            text: str

        Returns:
            number of tokens in the given text

        """
        num_tokens = len(self.tokenizer.encode(text))
        return num_tokens

    def trim_text_to_max_tokens(self, text: str) -> str:
        """
        Trims the text so that the number of tokens does not exceed max_tokens.

        Args:
        text (str): Original text string to be trimmed.

        Returns:
        str: Trimmed version of text or original text if under the limit.
        """
        # First check the number of tokens
        num_tokens = self.count_tokens(text)

        # If the number of tokens is within the limit, return the text as is
        if num_tokens <= self.max_tokens:
            return text

        # If the number exceeds the limit, trim the text
        # This is a simple trim, it may cut words in half; consider using word boundaries for a cleaner cut
        encoded_text = self.tokenizer.encode(text)
        trimmed_encoded_text = encoded_text[: self.max_tokens]
        # Decoding the trimmed text
        trimmed_text = self.tokenizer.decode(trimmed_encoded_text)
        return trimmed_text



import asyncio
import time
import os
from functools import lru_cache
from unittest.mock import patch
from cognee.shared.logging_utils import get_logger
from cognee.infrastructure.llm.rate_limiter import llm_rate_limiter
from cognee.infrastructure.llm.config import get_llm_config


async def test_rate_limiting_realistic():
    """
    Test the rate limiting feature with a smaller limit to demonstrate
    how rate limiting works in practice.
    """
    print("\n=== Testing Rate Limiting Feature (Realistic Test) ===")

    # Configure a lower rate limit for faster testing: 5 requests per 10 seconds
    os.environ["LLM_RATE_LIMIT_ENABLED"] = "true"
    os.environ["LLM_RATE_LIMIT_REQUESTS"] = "5"
    os.environ["LLM_RATE_LIMIT_INTERVAL"] = "10"

    # Clear the cached config and limiter
    get_llm_config.cache_clear()
    llm_rate_limiter._instance = None

    # Create fresh instances
    config = get_llm_config()
    print(
        f"Rate limit settings: {config.llm_rate_limit_enabled=}, {config.llm_rate_limit_requests=}, {config.llm_rate_limit_interval=}"
    )

    # We'll use monkey patching to guarantee rate limiting for test purposes
    with patch.object(llm_rate_limiter, "hit_limit") as mock_hit_limit:
        # Setup mock behavior: first 5 calls succeed, then one fails, then all succeed
        # This simulates the window moving after waiting
        mock_hit_limit.side_effect = [
            # First batch: 5 allowed, then 5 limited
            True,
            True,
            True,
            True,
            True,
            False,
            False,
            False,
            False,
            False,
            # Second batch after waiting: 2 allowed (partial window)
            True,
            True,
            False,
            False,
            False,
            # Third batch after full window reset: all 5 allowed
            True,
            True,
            True,
            True,
            True,
        ]

        limiter = llm_rate_limiter()
        print(f"Rate limiter initialized with {limiter._rate_per_minute} requests per minute")

        # First batch - should allow 5 and limit the rest
        print("\nBatch 1: Sending 10 requests (expecting only 5 to succeed)...")
        batch1_success = []
        batch1_failure = []

        for i in range(10):
            result = limiter.hit_limit()
            if result:
                batch1_success.append(i)
                print(f"✓ Request {i}: Success")
            else:
                batch1_failure.append(i)
                print(f"✗ Request {i}: Rate limited")

        print(f"Batch 1 results: {len(batch1_success)} successes, {len(batch1_failure)} failures")

        if len(batch1_failure) > 0:
            print(f"First rate-limited request: #{batch1_failure[0]}")

        # Wait for window to partially reset
        wait_time = 5  # seconds - half the rate limit interval
        print(f"\nWaiting for {wait_time} seconds to allow capacity to partially regenerate...")
        await asyncio.sleep(wait_time)

        # Second batch - should get some capacity back
        print("\nBatch 2: Sending 5 more requests (expecting 2 to succeed)...")
        batch2_success = []
        batch2_failure = []

        for i in range(5):
            result = limiter.hit_limit()
            if result:
                batch2_success.append(i)
                print(f"✓ Request {i}: Success")
            else:
                batch2_failure.append(i)
                print(f"✗ Request {i}: Rate limited")

        print(f"Batch 2 results: {len(batch2_success)} successes, {len(batch2_failure)} failures")

        # Wait for full window to reset
        full_wait = 10  # seconds - full rate limit interval
        print(f"\nWaiting for {full_wait} seconds for full capacity to regenerate...")
        await asyncio.sleep(full_wait)

        # Third batch - should have full capacity again
        print("\nBatch 3: Sending 5 requests (expecting all to succeed)...")
        batch3_success = []
        batch3_failure = []

        for i in range(5):
            result = limiter.hit_limit()
            if result:
                batch3_success.append(i)
                print(f"✓ Request {i}: Success")
            else:
                batch3_failure.append(i)
                print(f"✗ Request {i}: Rate limited")

        print(f"Batch 3 results: {len(batch3_success)} successes, {len(batch3_failure)} failures")

        # Calculate total successes and failures
        total_success = len(batch1_success) + len(batch2_success) + len(batch3_success)
        total_failure = len(batch1_failure) + len(batch2_failure) + len(batch3_failure)

        print(f"\nTotal requests: {total_success + total_failure}")
        print(f"Total successful: {total_success}")
        print(f"Total rate limited: {total_failure}")

        # Verify the rate limiting behavior
        if len(batch1_success) == 5 and len(batch1_failure) == 5:
            print("\n✅ PASS: Rate limiter correctly limited first batch to 5 requests")
        else:
            print(f"\n❌ FAIL: First batch should allow 5 requests, got {len(batch1_success)}")

        if len(batch2_success) == 2 and len(batch2_failure) == 3:
            print("✅ PASS: Rate limiter correctly allowed 2 requests after partial wait")
        else:
            print(f"❌ FAIL: Second batch should allow 2 requests, got {len(batch2_success)}")

        if len(batch3_success) == 5 and len(batch3_failure) == 0:
            print("✅ PASS: Rate limiter correctly allowed all requests after full window expired")
        else:
            print(f"❌ FAIL: Third batch should allow all 5 requests, got {len(batch3_success)}")

    print("=== Rate Limiting Test Complete ===\n")


async def main():
    """Run the realistic rate limiting test."""
    await test_rate_limiting_realistic()


if __name__ == "__main__":
    logger = get_logger()
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    try:
        loop.run_until_complete(main())
    finally:
        loop.run_until_complete(loop.shutdown_asyncgens())



"""
Rate limiter for LLM API calls.

This module provides rate limiting functionality for LLM API calls to prevent exceeding
API provider rate limits. The implementation uses the `limits` library with a moving window
strategy to limit requests.

Configuration is done through the LLMConfig with these settings:
- llm_rate_limit_enabled: Whether rate limiting is enabled (default: False)
- llm_rate_limit_requests: Maximum number of requests allowed per interval (default: 60)
- llm_rate_limit_interval: Interval in seconds for the rate limiting window (default: 60)

Usage:
1. Add the decorator to any function that makes API calls:
   @rate_limit_sync
   def my_function():
       # Function that makes API calls

2. For async functions, use the async decorator:
   @rate_limit_async
   async def my_async_function():
       # Async function that makes API calls

3. For automatic retrying on rate limit errors:
   @sleep_and_retry_sync
   def my_function():
       # Function that may experience rate limit errors

4. For async functions with automatic retrying:
   @sleep_and_retry_async
   async def my_async_function():
       # Async function that may experience rate limit errors

5. For embedding rate limiting (uses the same configuration but separate limiter):
   @embedding_rate_limit_async
   async def my_embedding_function():
       # Async function for embedding API calls

6. For embedding with auto-retry:
   @embedding_sleep_and_retry_async
   async def my_embedding_function():
       # Async function for embedding with auto-retry
"""

import time
import asyncio
import random
from functools import wraps
from limits import RateLimitItemPerMinute, storage
from limits.strategies import MovingWindowRateLimiter
from cognee.shared.logging_utils import get_logger
from cognee.infrastructure.llm.config import get_llm_config
import threading
import logging
import functools
import openai
import os

logger = get_logger()

# Common error patterns that indicate rate limiting
RATE_LIMIT_ERROR_PATTERNS = [
    "rate limit",
    "rate_limit",
    "ratelimit",
    "too many requests",
    "retry after",
    "capacity",
    "quota",
    "limit exceeded",
    "tps limit exceeded",
    "request limit exceeded",
    "maximum requests",
    "exceeded your current quota",
    "throttled",
    "throttling",
]

# Default retry settings
DEFAULT_MAX_RETRIES = 5
DEFAULT_INITIAL_BACKOFF = 1.0  # seconds
DEFAULT_BACKOFF_FACTOR = 2.0  # exponential backoff multiplier
DEFAULT_JITTER = 0.1  # 10% jitter to avoid thundering herd


class llm_rate_limiter:
    """
    Rate limiter for LLM API calls.

    This class implements a singleton pattern to ensure that rate limiting
    is consistent across all parts of the application. It uses the limits
    library with a moving window strategy to control request rates.

    The rate limiter converts the configured requests/interval to a per-minute
    rate for compatibility with the limits library's built-in rate limit items.
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(llm_rate_limiter, cls).__new__(cls)
            cls._instance._initialized = False
        return cls._instance

    def __init__(self):
        if self._initialized:
            return

        config = get_llm_config()
        self._enabled = config.llm_rate_limit_enabled
        self._requests = config.llm_rate_limit_requests
        self._interval = config.llm_rate_limit_interval

        # Using in-memory storage by default
        self._storage = storage.MemoryStorage()
        self._limiter = MovingWindowRateLimiter(self._storage)

        # Use the built-in per-minute rate limit item
        # We need to adjust the number of requests if interval isn't exactly 60s
        if self._interval == 60:
            self._rate_per_minute = self._requests
        else:
            self._rate_per_minute = int(self._requests * (60 / self._interval))

        self._initialized = True

        if self._enabled:
            logger.info(
                f"LLM Rate Limiter initialized: {self._requests} requests per {self._interval}s"
            )

    def hit_limit(self) -> bool:
        """
        Record a hit and check if limit is exceeded.

        This method checks whether making a request now would exceed the
        configured rate limit. If rate limiting is disabled, it always
        returns True.

        Returns:
            bool: True if the request is allowed, False otherwise.
        """
        if not self._enabled:
            return True

        # Create a fresh rate limit item for each check
        rate_limit = RateLimitItemPerMinute(self._rate_per_minute)

        # Use a consistent key for the API to ensure proper rate limiting
        return self._limiter.hit(rate_limit, "llm_api")

    def wait_if_needed(self) -> float:
        """
        Wait if rate limit is reached.

        This method blocks until the request can be made without exceeding
        the rate limit. It polls every 0.5 seconds.

        Returns:
            float: Time waited in seconds.
        """
        if not self._enabled:
            return 0

        waited = 0
        while not self.hit_limit():
            time.sleep(0.5)
            waited += 0.5

        return waited

    async def async_wait_if_needed(self) -> float:
        """
        Async wait if rate limit is reached.

        This method asynchronously waits until the request can be made without
        exceeding the rate limit. It polls every 0.5 seconds.

        Returns:
            float: Time waited in seconds.
        """
        if not self._enabled:
            return 0

        waited = 0
        while not self.hit_limit():
            await asyncio.sleep(0.5)
            waited += 0.5

        return waited


def rate_limit_sync(func):
    """
    Decorator for rate limiting synchronous functions.

    This decorator ensures that the decorated function respects the
    configured rate limits. If the rate limit would be exceeded,
    the decorator blocks until the request can be made.

    Args:
        func: The synchronous function to decorate.

    Returns:
        The decorated function.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        limiter = llm_rate_limiter()
        waited = limiter.wait_if_needed()
        if waited > 0:
            logger.debug(f"Rate limited LLM API call, waited for {waited}s")
        return func(*args, **kwargs)

    return wrapper


def rate_limit_async(func):
    """
    Decorator for rate limiting asynchronous functions.

    This decorator ensures that the decorated async function respects the
    configured rate limits. If the rate limit would be exceeded,
    the decorator asynchronously waits until the request can be made.

    Args:
        func: The asynchronous function to decorate.

    Returns:
        The decorated async function.
    """

    @wraps(func)
    async def wrapper(*args, **kwargs):
        limiter = llm_rate_limiter()
        waited = await limiter.async_wait_if_needed()
        if waited > 0:
            logger.debug(f"Rate limited LLM API call, waited for {waited}s")
        return await func(*args, **kwargs)

    return wrapper


def is_rate_limit_error(error):
    """
    Check if an error is related to rate limiting.

    Args:
        error: The exception to check.

    Returns:
        bool: True if the error is rate-limit related, False otherwise.
    """
    error_str = str(error).lower()
    return any(pattern.lower() in error_str for pattern in RATE_LIMIT_ERROR_PATTERNS)


def calculate_backoff(
    attempt,
    initial_backoff=DEFAULT_INITIAL_BACKOFF,
    backoff_factor=DEFAULT_BACKOFF_FACTOR,
    jitter=DEFAULT_JITTER,
):
    """
    Calculate the backoff time for a retry attempt with jitter.

    Args:
        attempt: The current retry attempt (0-based).
        initial_backoff: The initial backoff time in seconds.
        backoff_factor: The multiplier for exponential backoff.
        jitter: The jitter factor to avoid thundering herd.

    Returns:
        float: The backoff time in seconds.
    """
    backoff = initial_backoff * (backoff_factor**attempt)
    jitter_amount = backoff * jitter
    return backoff + random.uniform(-jitter_amount, jitter_amount)


def sleep_and_retry_sync(
    max_retries=DEFAULT_MAX_RETRIES,
    initial_backoff=DEFAULT_INITIAL_BACKOFF,
    backoff_factor=DEFAULT_BACKOFF_FACTOR,
    jitter=DEFAULT_JITTER,
):
    """
    Decorator that automatically retries a synchronous function when rate limit errors occur.

    This decorator implements an exponential backoff strategy with jitter
    to handle rate limit errors efficiently.

    Args:
        max_retries: Maximum number of retry attempts.
        initial_backoff: Initial backoff time in seconds.
        backoff_factor: Multiplier for exponential backoff.
        jitter: Jitter factor to avoid the thundering herd problem.

    Returns:
        The decorated function.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            attempt = 0
            while True:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    attempt += 1
                    if not is_rate_limit_error(e) or attempt > max_retries:
                        raise

                    backoff_time = calculate_backoff(
                        attempt, initial_backoff, backoff_factor, jitter
                    )
                    logger.warning(
                        f"Rate limit hit, retrying in {backoff_time:.2f}s "
                        f"(attempt {attempt}/{max_retries}): {str(e)}"
                    )
                    time.sleep(backoff_time)

        return wrapper

    return decorator


def sleep_and_retry_async(
    max_retries=DEFAULT_MAX_RETRIES,
    initial_backoff=DEFAULT_INITIAL_BACKOFF,
    backoff_factor=DEFAULT_BACKOFF_FACTOR,
    jitter=DEFAULT_JITTER,
):
    """
    Decorator that automatically retries an asynchronous function when rate limit errors occur.

    This decorator implements an exponential backoff strategy with jitter
    to handle rate limit errors efficiently.

    Args:
        max_retries: Maximum number of retry attempts.
        initial_backoff: Initial backoff time in seconds.
        backoff_factor: Multiplier for exponential backoff.
        jitter: Jitter factor to avoid the thundering herd problem.

    Returns:
        The decorated async function.
    """

    def decorator(func):
        @wraps(func)
        async def wrapper(*args, **kwargs):
            attempt = 0
            while True:
                try:
                    return await func(*args, **kwargs)
                except Exception as e:
                    attempt += 1
                    if not is_rate_limit_error(e) or attempt > max_retries:
                        raise

                    backoff_time = calculate_backoff(
                        attempt, initial_backoff, backoff_factor, jitter
                    )
                    logger.warning(
                        f"Rate limit hit, retrying in {backoff_time:.2f}s "
                        f"(attempt {attempt}/{max_retries}): {str(e)}"
                    )
                    await asyncio.sleep(backoff_time)

        return wrapper

    return decorator



from datetime import datetime, timezone
from sqlalchemy.sql import func

from sqlalchemy import Column, DateTime, Float, Integer, JSON, UUID

from cognee.infrastructure.databases.relational import Base
from uuid import uuid4


class GraphMetrics(Base):
    __tablename__ = "graph_metrics"

    # TODO: Change ID to reflect unique id of graph database
    id = Column(UUID, primary_key=True, default=uuid4)
    num_tokens = Column(Integer, nullable=True)
    num_nodes = Column(Integer, nullable=True)
    num_edges = Column(Integer, nullable=True)
    mean_degree = Column(Float, nullable=True)
    edge_density = Column(Float, nullable=True)
    num_connected_components = Column(Integer, nullable=True)
    sizes_of_connected_components = Column(JSON, nullable=True)
    num_selfloops = Column(Integer, nullable=True)
    diameter = Column(Integer, nullable=True)
    avg_shortest_path_length = Column(Float, nullable=True)
    avg_clustering = Column(Float, nullable=True)

    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(DateTime(timezone=True), onupdate=func.now())



from .create_user import create_user
from cognee.base_config import get_base_config


async def create_default_user():
    base_config = get_base_config()
    default_user_email = base_config.default_user_email or "default_user@example.com"
    default_user_password = base_config.default_user_password or "default_password"

    user = await create_user(
        email=default_user_email,
        password=default_user_password,
        is_superuser=True,
        is_active=True,
        is_verified=True,
        auto_login=True,
    )

    return user



import pytest
from typing import List, Optional
from cognee.infrastructure.engine import DataPoint

from cognee.modules.graph.utils import get_graph_from_model


class Document(DataPoint):
    path: str
    metadata: dict = {"index_fields": []}


class DocumentChunk(DataPoint):
    part_of: Document
    text: str
    contains: List["Entity"] = None
    metadata: dict = {"index_fields": ["text"]}


class EntityType(DataPoint):
    name: str
    metadata: dict = {"index_fields": ["name"]}


class Entity(DataPoint):
    name: str
    is_type: EntityType
    metadata: dict = {"index_fields": ["name"]}


DocumentChunk.model_rebuild()


@pytest.mark.asyncio
async def test_get_graph_from_model_simple_structure():
    """Tests simple pydantic structure for get_graph_from_model"""

    entitytype = EntityType(
        name="TestType",
    )

    entity = Entity(name="TestEntity", is_type=entitytype)

    added_nodes = {}
    added_edges = {}
    visited_properties = {}

    nodes, edges = await get_graph_from_model(entity, added_nodes, added_edges, visited_properties)

    assert len(nodes) == 2, f"Expected 2 nodes, got {len(nodes)}"
    assert len(edges) == 1, f"Expected 1 edges, got {len(edges)}"

    edge_key = str(entity.id) + str(entitytype.id) + "is_type"
    assert edge_key in added_edges, f"Edge {edge_key} not found"


@pytest.mark.asyncio
async def test_get_graph_from_model_with_document_and_chunk():
    """Tests multiple entities to document connection"""
    doc = Document(path="test/path")
    doc_chunk = DocumentChunk(part_of=doc, text="This is a chunk of text", contains=[])
    entity_type = EntityType(name="Person")
    entity = Entity(name="Alice", is_type=entity_type)
    entity2 = Entity(name="Alice2", is_type=entity_type)
    doc_chunk.contains.append(entity)
    doc_chunk.contains.append(entity2)

    added_nodes = {}
    added_edges = {}
    visited_properties = {}

    nodes, edges = await get_graph_from_model(
        doc_chunk, added_nodes, added_edges, visited_properties
    )

    assert len(nodes) == 5, f"Expected 5 nodes, got {len(nodes)}"
    assert len(edges) == 5, f"Expected 5 edges, got {len(edges)}"


@pytest.mark.asyncio
async def test_get_graph_from_model_duplicate_references():
    """Tests duplicated objects in document list"""
    doc = Document(path="test/path")
    doc_chunk = DocumentChunk(part_of=doc, text="Chunk with duplicates", contains=[])

    entity_type = EntityType(name="Animal")
    shared_entity = Entity(name="Cat", is_type=entity_type)

    doc_chunk.contains.extend([shared_entity, shared_entity, shared_entity])

    added_nodes = {}
    added_edges = {}
    visited_properties = {}

    nodes, edges = await get_graph_from_model(
        doc_chunk, added_nodes, added_edges, visited_properties
    )

    assert len(nodes) == 4, f"Expected 4 nodes, got {len(nodes)}"
    assert len(edges) == 3, f"Expected 3 edges, got {len(edges)}"


@pytest.mark.asyncio
async def test_get_graph_from_model_multi_level_nesting():
    """Tests multi level nested structure extraction"""
    doc = Document(path="multi-level/path")

    chunk1 = DocumentChunk(part_of=doc, text="Chunk 1 text", contains=[])
    chunk2 = DocumentChunk(part_of=doc, text="Chunk 2 text", contains=[])

    entity_type_vehicle = EntityType(name="Vehicle")
    entity_type_person = EntityType(name="Person")

    entity_car = Entity(name="Car", is_type=entity_type_vehicle)
    entity_bike = Entity(name="Bike", is_type=entity_type_vehicle)
    entity_alice = Entity(name="Alice", is_type=entity_type_person)

    chunk1.contains.extend([entity_car, entity_bike])
    chunk2.contains.append(entity_alice)

    added_nodes = {}
    added_edges = {}
    visited_properties = {}

    nodes, edges = await get_graph_from_model(chunk1, added_nodes, added_edges, visited_properties)

    nodes2, edges2 = await get_graph_from_model(
        chunk2, added_nodes, added_edges, visited_properties
    )

    all_nodes = nodes + nodes2
    all_edges = edges + edges2

    assert len(all_nodes) == 8, f"Expected 8 nodes, got {len(all_nodes)}"
    assert len(all_edges) == 8, f"Expected 8 edges, got {len(all_edges)}"


@pytest.mark.asyncio
async def test_get_graph_from_model_no_contains():
    """Tests graph from model with empty contains element"""
    doc = Document(path="empty-contains/path")
    chunk = DocumentChunk(part_of=doc, text="A chunk with no entities", contains=[])

    added_nodes = {}
    added_edges = {}
    visited_properties = {}

    nodes, edges = await get_graph_from_model(chunk, added_nodes, added_edges, visited_properties)

    assert len(nodes) == 2, f"Expected 2 nodes, got {len(nodes)}"
    assert len(edges) == 1, f"Expected 1 edge, got {len(edges)}"



from deepeval.test_case import LLMTestCase
from typing import Optional


class ExactMatchMetric:
    def __init__(self) -> None:
        self.score: Optional[float] = None
        self.reason: Optional[str] = None

    def measure(self, test_case: "LLMTestCase") -> float:
        actual = test_case.actual_output.strip().lower() if test_case.actual_output else ""
        expected = test_case.expected_output.strip().lower() if test_case.expected_output else ""
        self.score = 1.0 if actual == expected else 0.0
        self.reason = "Exact match" if self.score == 1.0 else "Not an exact match"
        return self.score



import os
from cognee.shared.logging_utils import get_logger
import pathlib
import cognee

from cognee.modules.data.models import Data
from cognee.modules.users.methods import get_default_user
from cognee.modules.search.types import SearchType
from cognee.modules.search.operations import get_history

logger = get_logger()


async def test_local_file_deletion(data_text, file_location):
    from sqlalchemy import select
    import hashlib
    from cognee.infrastructure.databases.relational import get_relational_engine

    engine = get_relational_engine()

    async with engine.get_async_session() as session:
        # Get hash of data contents
        encoded_text = data_text.encode("utf-8")
        data_hash = hashlib.md5(encoded_text).hexdigest()
        # Get data entry from database based on hash contents
        data = (await session.scalars(select(Data).where(Data.content_hash == data_hash))).one()
        assert os.path.isfile(data.raw_data_location), (
            f"Data location doesn't exist: {data.raw_data_location}"
        )
        # Test deletion of data along with local files created by cognee
        await engine.delete_data_entity(data.id)
        assert not os.path.exists(data.raw_data_location), (
            f"Data location still exists after deletion: {data.raw_data_location}"
        )

    async with engine.get_async_session() as session:
        # Get data entry from database based on file path
        data = (
            await session.scalars(select(Data).where(Data.raw_data_location == file_location))
        ).one()
        assert os.path.isfile(data.raw_data_location), (
            f"Data location doesn't exist: {data.raw_data_location}"
        )
        # Test local files not created by cognee won't get deleted
        await engine.delete_data_entity(data.id)
        assert os.path.exists(data.raw_data_location), (
            f"Data location doesn't exists: {data.raw_data_location}"
        )


async def test_getting_of_documents(dataset_name_1):
    # Test getting of documents for search per dataset
    from cognee.modules.users.permissions.methods import get_document_ids_for_user

    user = await get_default_user()
    document_ids = await get_document_ids_for_user(user.id, [dataset_name_1])
    assert len(document_ids) == 1, (
        f"Number of expected documents doesn't match {len(document_ids)} != 1"
    )

    # Test getting of documents for search when no dataset is provided
    user = await get_default_user()
    document_ids = await get_document_ids_for_user(user.id)
    assert len(document_ids) == 2, (
        f"Number of expected documents doesn't match {len(document_ids)} != 2"
    )


async def main():
    cognee.config.set_vector_db_config(
        {
            "vector_db_url": "http://localhost:3002",
            "vector_db_key": "test-token",
            "vector_db_provider": "chromadb",
        }
    )

    data_directory_path = str(
        pathlib.Path(
            os.path.join(pathlib.Path(__file__).parent, ".data_storage/test_chromadb")
        ).resolve()
    )
    cognee.config.data_root_directory(data_directory_path)
    cognee_directory_path = str(
        pathlib.Path(
            os.path.join(pathlib.Path(__file__).parent, ".cognee_system/test_chromadb")
        ).resolve()
    )
    cognee.config.system_root_directory(cognee_directory_path)

    await cognee.prune.prune_data()
    await cognee.prune.prune_system(metadata=True)

    dataset_name_1 = "natural_language"
    dataset_name_2 = "quantum"

    explanation_file_path = os.path.join(
        pathlib.Path(__file__).parent, "test_data/Natural_language_processing.txt"
    )
    await cognee.add([explanation_file_path], dataset_name_1)

    text = """A quantum computer is a computer that takes advantage of quantum mechanical phenomena.
    At small scales, physical matter exhibits properties of both particles and waves, and quantum computing leverages this behavior, specifically quantum superposition and entanglement, using specialized hardware that supports the preparation and manipulation of quantum states.
    Classical physics cannot explain the operation of these quantum devices, and a scalable quantum computer could perform some calculations exponentially faster (with respect to input size scaling) than any modern "classical" computer. In particular, a large-scale quantum computer could break widely used encryption schemes and aid physicists in performing physical simulations; however, the current state of the technology is largely experimental and impractical, with several obstacles to useful applications. Moreover, scalable quantum computers do not hold promise for many practical tasks, and for many important tasks quantum speedups are proven impossible.
    The basic unit of information in quantum computing is the qubit, similar to the bit in traditional digital electronics. Unlike a classical bit, a qubit can exist in a superposition of its two "basis" states. When measuring a qubit, the result is a probabilistic output of a classical bit, therefore making quantum computers nondeterministic in general. If a quantum computer manipulates the qubit in a particular way, wave interference effects can amplify the desired measurement results. The design of quantum algorithms involves creating procedures that allow a quantum computer to perform calculations efficiently and quickly.
    Physically engineering high-quality qubits has proven challenging. If a physical qubit is not sufficiently isolated from its environment, it suffers from quantum decoherence, introducing noise into calculations. Paradoxically, perfectly isolating qubits is also undesirable because quantum computations typically need to initialize qubits, perform controlled qubit interactions, and measure the resulting quantum states. Each of those operations introduces errors and suffers from noise, and such inaccuracies accumulate.
    In principle, a non-quantum (classical) computer can solve the same computational problems as a quantum computer, given enough time. Quantum advantage comes in the form of time complexity rather than computability, and quantum complexity theory shows that some quantum algorithms for carefully selected tasks require exponentially fewer computational steps than the best known non-quantum algorithms. Such tasks can in theory be solved on a large-scale quantum computer whereas classical computers would not finish computations in any reasonable amount of time. However, quantum speedup is not universal or even typical across computational tasks, since basic tasks such as sorting are proven to not allow any asymptotic quantum speedup. Claims of quantum supremacy have drawn significant attention to the discipline, but are demonstrated on contrived tasks, while near-term practical use cases remain limited.
    """

    await cognee.add([text], dataset_name_2)

    await cognee.cognify([dataset_name_2, dataset_name_1])

    from cognee.infrastructure.databases.vector import get_vector_engine

    await test_getting_of_documents(dataset_name_1)

    vector_engine = get_vector_engine()
    random_node = (await vector_engine.search("Entity_name", "Quantum computer"))[0]
    random_node_name = random_node.payload["text"]

    search_results = await cognee.search(
        query_type=SearchType.INSIGHTS, query_text=random_node_name
    )
    assert len(search_results) != 0, "The search results list is empty."
    print("\n\nExtracted sentences are:\n")
    for result in search_results:
        print(f"{result}\n")

    search_results = await cognee.search(
        query_type=SearchType.CHUNKS, query_text=random_node_name, datasets=[dataset_name_2]
    )
    assert len(search_results) != 0, "The search results list is empty."
    print("\n\nExtracted chunks are:\n")
    for result in search_results:
        print(f"{result}\n")

    graph_completion = await cognee.search(
        query_type=SearchType.GRAPH_COMPLETION,
        query_text=random_node_name,
        datasets=[dataset_name_2],
    )
    assert len(graph_completion) != 0, "Completion result is empty."
    print("Completion result is:")
    print(graph_completion)

    search_results = await cognee.search(
        query_type=SearchType.SUMMARIES, query_text=random_node_name
    )
    assert len(search_results) != 0, "Query related summaries don't exist."
    print("\n\nExtracted summaries are:\n")
    for result in search_results:
        print(f"{result}\n")

    user = await get_default_user()
    history = await get_history(user.id)
    assert len(history) == 8, "Search history is not correct."

    await cognee.prune.prune_data()
    assert not os.path.isdir(data_directory_path), "Local data files are not deleted"

    await cognee.prune.prune_system(metadata=True)
    tables_in_database = await vector_engine.get_collection_names()
    assert len(tables_in_database) == 0, "ChromaDB database is not empty"


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())



from cognee.infrastructure.engine import DataPoint
from cognee.modules.engine.models.EntityType import EntityType
from typing import Optional


class Entity(DataPoint):
    name: str
    is_a: Optional[EntityType] = None
    description: str

    metadata: dict = {"index_fields": ["name"]}
